import re
import os
import numpy as np
from math import sqrt
from pychemia.utils.netcdf import file2dict
from ..codes import CodeOutput


class AbinitOutput(CodeOutput):

    def __init__(self, filename='abinit.out'):

        CodeOutput.__init__(self)
        self.filename = None
        self.data = ''
        if os.path.isfile(filename):
            self.filename = filename
        elif os.path.isdir(filename) and os.path.isfile(filename + os.sep + 'abinit.out'):
            self.filename = filename + os.sep + 'abinit.out'
        if self.filename is not None:
            rf = open(self.filename)
            self.data = rf.read()
            if "Calculation completed." in self.data:
                self.read()

    def reload(self):
        rf = open(self.filename)
        self.data = rf.read()

    def get_energetics(self):

        ret = re.findall('ETOT\s*([\d]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)\s*([.E\d\-+]+)',
                         self.data)
        if not ret:
            raise RuntimeError("Could not extract energetic lines from: %s" % self.filename)

        newret = []
        for y in ret:
            try:
                etotline = [float(x) for x in y]
                newret.append(etotline)
            except ValueError:
                print("This line could not be parsed correctly: %s" % y)

        ret = np.array(newret)
        energetics = {'iter': list(ret[:, 0]),
                      'etot': list(ret[:, 1]),
                      'deltaEh': list(ret[:, 2]),
                      'residm': list(ret[:, 3]),
                      'nres2': list(ret[:, 4])}
        return energetics

    def read(self):
        self.output_values = self.get_energetics()

    @property
    def is_finished(self):
        return "Calculation completed." in self.data

    def get_occupation_matrices(self):

        occ_block = re.findall(r"=== For Atom[\s\w\d\-.,=>:]*\n", self.data)
        if not occ_block:
            raise RuntimeError("ERROR: Could not get occupation matrices blocks: %s" % self.filename)
        if len(occ_block) != 1:
            raise RuntimeError("ERROR: Occupation matrices should be in one block from the output")

        occs = re.findall('Occupation matrix for spin[\s\d\w]*([\s\d\-.]*)', occ_block[0])
        atoms = [int(x) for x in re.findall('For Atom([\s\w\d]*)', occ_block[0])]
        spins = [int(x) for x in re.findall('Occupation matrix for spin([\s\w\d]*\n)', occ_block[0])]

        # print(occs)
        # print(atoms)
        # print(spins)

        # print("Number of atoms: %d" % len(atoms))
        # print("Number of spins blocks: %d" % len(spins))
        # print("Number of occ matrices: %d" % len(occs))

        if 2 * len(atoms) != len(spins) or len(spins) != len(occs):
            raise RuntimeError("Number of matrices is not consistent with number of atoms and spin")

        ret = {}
        for i in range(len(occs)):
            atom = atoms[int((i + 1) / 2) + (i + 1) % 2 - 1]
            spin = spins[i]
            occ_matrix = [float(x) for x in occs[i].strip().split()]
            if len(occ_matrix) not in [25, 49]:
                raise RuntimeError("Occupation matrices must be 5x5 or 7x7")
            ret[(atom, spin)] = np.array(occ_matrix).reshape((-1, int(sqrt(len(occ_matrix))),
                                                              int(sqrt(len(occ_matrix)))))
        return ret

    def get_dmatpawu(self, both_spins=True):
        om = self.get_occupation_matrices()
        min_atom = min([x[0] for x in om.keys()])
        max_atom = max([x[0] for x in om.keys()])
        min_spin = min([x[1] for x in om.keys()])
        max_spin = max([x[1] for x in om.keys()])
        if not both_spins:
            max_spin = min_spin
        ret = None
        for i in range(min_atom, max_atom + 1):
            for j in range(min_spin, max_spin + 1):
                if ret is None:
                    ret = om[(i, j)]
                else:
                    ret = np.vstack((ret, om[(i, j)]))
        return ret

    def get_integrated_electronic_densities(self):
        """
        Return the integrated electronic densities in atomic spheres from the output
        Data is returned serialized as python dictionary with keys
        'radius', 'up_density', 'dn_density', 'total_up+dn' and 'diff_up-dn'
        If the method is unable to parse the output file None is returned.
        
        :return: (dict) Integrated electronic densities around an atomic radius, up, down, sum and diff
        """

        ref = re.findall("Integrated electronic and magnetization densities in atomic spheres:\s*"
                         "---------------------------------------------------------------------\s*"
                         "Note: Diff\(up-dn\) is a rough approximation of local magnetic moment\s*"
                         "Atom\s*Radius\s*up_density\s*dn_density\s*Total\(up\+dn\)\s*Diff\(up-dn\)\s*"
                         "([\s\d\w.-]*)\n -+\n", self.data)
        # Return None if could not parse the magnetization block
        if len(ref) == 0:
            return None
        else:
            magmoms = np.fromstring(ref[0], dtype=float, sep=' ')
            magmoms.reshape((-1, 6))
            return {'radius': list(magmoms[:, 1]),
                    'up_density': list(magmoms[:, 2]),
                    'dn_density': list(magmoms[:, 3]),
                    'total_up+dn': list(magmoms[:, 4]),
                    'diff_up-dn': list(magmoms[:, 5])}

    @staticmethod
    def read_output_netcdf(filename):
        return file2dict(filename)

# DEPRECATED CODE, all ABINIT specific operations from orbitaldftu
# is moved here. 
#

# def get_final_correlation_matrices_from_output(filename):
#     rf = open(filename)
#     data = rf.read()
#     mainblock = re.findall('LDA\+U DATA[\s\w\d\-.=,>:]*\n', data)
#     if len(mainblock) != 1:
#         for isection in mainblock:
#             print(isection)
#         raise ValueError("Occupations for correlated orbitals could not be detected right")


#     pattern = "For Atom\s*(\d+), occupations for correlated orbitals. lpawu =\s*([\d]+)\s*Atom\s*[\d]+\s*. Occ. " \
#               "for lpawu and for spin\s*\d+\s*=\s*([\d\.]+)\s*Atom\s*[\d]+\s*. Occ. for lpawu and for " \
#               "spin\s*\d+\s*=\s*([\d\.]+)\s*=> On atom\s*\d+\s*,  local Mag. for lpawu is[\s\d\w\.\-]*== Occupation " \
#               "matrix for correlated orbitals:\s*Occupation matrix for spin  1\s*([\d\.\-\s]*)Occupation matrix " \
#               "for spin  2\s*([\d\.\-\s]*)"
#     ans = re.findall(pattern, mainblock[0])
#     # print(ans)

#     ret = []
#     for i in ans:
#         atom_data = {'atom number': int(i[0]),
#                      'orbital': int(i[1]),
#                      'occ spin 1': float(i[2]),
#                      'occ spin 2': float(i[3])}
#         matrix = [float(x) for x in i[4].split()]
#         atom_data['matrix spin 1'] = list(matrix)
#         matrix = [float(x) for x in i[5].split()]
#         atom_data['matrix spin 2'] = list(matrix)
#         ret.append(atom_data)
#     return ret


# def get_final_dmatpawu(filename):
#     ret = get_final_correlation_matrices_from_output(filename)
#     # Looking for input file
#     dirname = os.path.dirname(os.path.abspath(filename))
#     outnc = [ x for x in os.listdir(dirname) if x[-6:]=='OUT.nc']
#     if len(outnc) < 1:
#         raise ValueError("ERROR: Could not find *.OUT.nc needed to determine nspol, nspden and nspinor")
#     variables = netcdf2dict(dirname + os.sep + outnc[0])

#     if 'nspinor' in variables:
#         nspinor = variables['nspinor']
#     else:
#         nspinor = 1

#     if 'nsppol' in variables:
#         nsppol =  variables['nsppol']
#     else:
#         nsppol = 1

#     if 'nspden' in variables:
#         nspden = variables['nspden']
#     else:
#         nspden = nsppol

#     # number of matrices per atom:
#     nmpa = get_num_matrices_per_atom(nsppol, nspden, nspinor)

#     dmatpawu = []
#     for i in ret:
#         dmatpawu += i['matrix spin 1']
#         if nmpa == 2:
#             dmatpawu += i['matrix spin 2']        
#     return dmatpawu



from mayavi import mlab
from pychemia.utils.periodic import covalent_radius
import numpy as np


class StructurePlot:

    def __init__(self, structure):
        """

        :param structure: (Structure) A pychemia Structure object
        """
        self.structure = structure

    def plot(self, figname=None, size=(300, 325), view=(30, 30), color=(1.0, 1.0, 1.0)):

        fig = mlab.figure(size=size)
        figure = mlab.gcf()
        fig.scene.disable_render = True
        figure.scene.background = (0.0, 0.0, 0.0)
        mlab.view(0, 90, distance=0.2)
        assert (self.structure.natom > 0)

        x = self.structure.positions[:, 0]
        y = self.structure.positions[:, 1]
        z = self.structure.positions[:, 2]
        cr = covalent_radius(self.structure.symbols)
        s = np.apply_along_axis(np.linalg.norm, 1, self.structure.positions)

        mlab.points3d(x, y, z, s, scale_factor=1.0, resolution=8, opacity=1.0,
                      color=color,
                      scale_mode='none')

        if self.structure.is_crystal:
            frame, line1, line2, line3 = self.structure.get_cell().get_path()

            mlab.plot3d(frame[:, 0], frame[:, 1], frame[:, 2], tube_radius=.02, color=(1, 1, 1))
            mlab.plot3d(line1[:, 0], line1[:, 1], line1[:, 2], tube_radius=.02, color=(1, 1, 1))
            mlab.plot3d(line2[:, 0], line2[:, 1], line2[:, 2], tube_radius=.02, color=(1, 1, 1))
            mlab.plot3d(line3[:, 0], line3[:, 1], line3[:, 2], tube_radius=.02, color=(1, 1, 1))
        else:
            for i in range(self.structure.natom - 1):
                for j in range(i + 1, self.structure.natom):
                    vector = self.structure.positions[i] - self.structure.positions[j]
                    mvector = np.linalg.norm(vector)
                    uvector = 1.0 / mvector * vector
                    if 2 * mvector < covalent_radius(self.structure.symbols[i]) + \
                            covalent_radius(self.structure.symbols[j]):
                        pair = np.concatenate(
                            (self.structure.positions[i] - 0.1 * uvector,
                             self.structure.positions[j] + 0.1 * uvector)).reshape((-1, 3))
                        mlab.plot3d(pair[:, 0], pair[:, 1], pair[:, 2], tube_radius=0.15, opacity=1.0, color=(1, 1, 1))

        mlab.view(distance=12.0)
        fig.scene.disable_render = False
        if figname is not None:
            mlab.savefig(figname)
        return figure




import math
import sys
import numpy as np
import scipy.spatial
from pychemia import Composition, Structure, pcm_log
from pychemia.analysis import ClusterAnalysis, ClusterMatch
from pychemia.code.lennardjones import lj_compact_evaluate
from pychemia.utils.mathematics import unit_vector, length_vectors, unit_vectors, rotate_towards_axis, length_vector
from pychemia.utils.periodic import covalent_radius, atomic_number
from pychemia.utils.serializer import generic_serializer
from pychemia.code.lennardjones import LennardJones
from pychemia.external.symmol import get_point_group
from ._population import Population
from ._distances import FingerPrints, StructureDistances


class LJCluster(Population):

    def __init__(self, name, composition=None, tag='global', target_forces=1E-3, value_tol=1E-2,
                 distance_tolerance=0.1, minimal_density=70.0, refine=True, direct_evaluation=False):
        if composition is not None:
            self.composition = Composition(composition)
        else:
            self.composition = None
        Population.__init__(self, name=name, tag=tag, direct_evaluation=direct_evaluation,
                            distance_tolerance=distance_tolerance)
        self.tag = tag
        self.target_forces = target_forces
        self.value_tol = value_tol
        self.minimal_density = minimal_density
        self.refine = refine
        self.fingerprinter = FingerPrints(self.pcdb)
        self.distancer = StructureDistances(self.pcdb)

    def add_random(self):
        """
        Add one random structure to the population
        """
        if self.composition is None:
            raise ValueError('No composition associated to this population')
        comp = self.composition.composition.copy()
        structure = Structure.random_cluster(composition=comp)

        return self.new_entry(structure), None

    def get_duplicates(self, ids, tolerance=None, fast=True):
        ret = {}
        selection = self.ids_sorted(ids)
        values = np.array([self.value(i) for i in selection])
        if len(values) == 0:
            return ret
        diffs = np.ediff1d(values)

        for i in range(len(diffs)):
            idiff = diffs[i]
            if idiff < self.value_tol:
                ident1 = selection[i]
                ident2 = selection[i + 1]
                pcm_log.debug('Testing distances between %s and %s' % (str(ident1), str(ident2)))
                distance = self.distance(ident1, ident2)
                if distance < self.distance_tolerance:
                    pcm_log.debug('Distance %7.3f < %7.3f' % (distance, self.distance_tolerance))
                    ret[ident2] = ident1
        if len(ret) > 0:
            pcm_log.debug('Number of duplicates %d' % len(ret))
        return ret

    def cross(self, ids):
        if len(ids) != 2:
            raise ValueError("Crossing only implemented between two clusters")

        entry0 = self.get_entry(ids[0])
        entry1 = self.get_entry(ids[1])

        pos0 = np.array(entry0['structure']['positions']).reshape((-1, 3))
        pos1 = np.array(entry1['structure']['positions']).reshape((-1, 3))

        cut = np.random.randint(1, len(pos0))

        new_pos0 = np.concatenate((pos0[:cut], pos1[cut:]))
        new_pos1 = np.concatenate((pos1[:cut], pos0[cut:]))

        new_structure = Structure(positions=new_pos0, symbols=entry0['structure']['symbols'],
                                  periodicity=False)
        entry_id = self.new_entry(structure=new_structure)
        new_structure = Structure(positions=new_pos1, symbols=entry0['structure']['symbols'],
                                  periodicity=False)
        entry_jd = self.new_entry(structure=new_structure)

        return entry_id, entry_jd

    def distance(self, entry_id, entry_jd, rcut=50):
        """
        Return a measure of the distance between two clusters by computing
        a n-dimensional vector of the distances between each atom to the
        origin and

        :param rcut:
        :param entry_id: The id of one population entry
        :param entry_jd: The id of another population entry
        :return: (int) The distance between two clusters
        """

        ids_pair = tuple(np.sort([entry_id, entry_jd]))
        distance_entry = self.distancer.get_distance(ids_pair)

        if distance_entry is None:
            fingerprints = {}
            for entry_ijd in [entry_id, entry_jd]:

                if self.fingerprinter.get_fingerprint(entry_ijd) is None:
                    structure = self.get_structure(entry_ijd)
                    analysis = ClusterAnalysis(structure)
                    x, ys = analysis.discrete_radial_distribution_function()
                    fingerprint = {'_id': entry_ijd}
                    for k in ys:
                        atomic_number1 = atomic_number(k[0])
                        atomic_number2 = atomic_number(k[1])
                        pair = '%06d' % min(atomic_number1 * 1000 + atomic_number2,
                                            atomic_number2 * 1000 + atomic_number1)
                        fingerprint[pair] = list(ys[k])

                    if self.fingerprinter.get_fingerprint(entry_ijd) is None:
                        self.fingerprinter.set_fingerprint(fingerprint)
                    else:
                        self.fingerprinter.update(entry_ijd, fingerprint)
                    fingerprints[entry_ijd] = fingerprint
                else:
                    fingerprints[entry_ijd] = self.fingerprinter.get_fingerprint(entry_ijd)

            dij = []
            for pair in fingerprints[entry_id]:
                if pair in fingerprints[entry_jd] and pair != '_id':
                    vect1 = fingerprints[entry_id][pair]
                    vect2 = fingerprints[entry_jd][pair]
                    if len(vect1) < len(vect2):
                        tmp = np.zeros(len(vect2))
                        tmp[:len(vect1)] = vect1
                        vect1 = tmp
                    elif len(vect1) > len(vect2):
                        tmp = np.zeros(len(vect1))
                        tmp[:len(vect2)] = vect2
                        vect2 = tmp
                    uvect1 = unit_vector(vect1)
                    uvect2 = unit_vector(vect2)
                    dij.append(0.5 * (1.0 - np.dot(uvect1, uvect2)))
            distance = float(np.mean(dij))
            self.distancer.set_distance(ids_pair, distance)
        else:
            distance = distance_entry['distance']
        return distance

    @property
    def to_dict(self):
        return {'name': self.name,
                'tag': self.tag,
                'target_forces': self.target_forces,
                'value_tol': self.value_tol,
                'distance_tolerance': self.distance_tolerance,
                'minimal_density': self.minimal_density}

    def get2_duplicates(self, ids, fast=True):
        dupes_dict = {}
        dupes_list = []
        selection = self.ids_sorted(ids)
        pcm_log.debug('Searching duplicates in %d structures' % len(selection))
        for i in range(len(selection) - 1):
            ncomps = 0
            entry_id = selection[i]
            if fast and entry_id in dupes_list:
                continue
            sys.stdout.write(" %5d of %5d: " % (i, len(selection)))
            value_i = self.value(entry_id)
            for j in range(i + 1, len(selection)):
                entry_jd = selection[j]
                if fast and entry_jd in dupes_list:
                    continue
                value_j = self.value(entry_jd)
                if abs(value_i - value_j) < self.value_tol:
                    ncomps += 1
                    distance = self.distance(entry_id, entry_jd)
                    if distance < self.distance_tolerance:
                        if entry_id in dupes_dict:
                            dupes_dict[entry_id].append(entry_jd)
                        else:
                            if fast:
                                dupes_dict[entry_id] = [entry_jd]
                            else:
                                dupes_dict[entry_id] = entry_jd
                        dupes_list.append(entry_jd)
            sys.stdout.write(' comparisons: %d\n' % ncomps)
        return dupes_dict, [x for x in selection if x in dupes_list]

    def is_evaluated(self, entry_id):

        entry = self.get_entry(entry_id)
        if entry is not None and 'properties' not in entry:
            raise ValueError('Anomalous condition for %s' % entry_id)

        if entry is not None and entry['properties'] is not None:
            properties = entry['properties']
            if 'forces' not in properties:
                forces = None
            elif properties['forces'] is None:
                forces = None
            else:
                forces = np.max(np.apply_along_axis(np.linalg.norm, 1, np.array(properties['forces']).reshape((-1, 3))))
        else:
            forces = None

        if forces is not None and forces < self.target_forces:
            return True
        else:
            return False

    def from_dict(self, population_dict):
        return LJCluster(name=population_dict['name'],
                         tag=population_dict['tag'],
                         target_forces=population_dict['target_forces'],
                         value_tol=population_dict['value_tol'],
                         distance_tolerance=population_dict['distance_tolerance'],
                         minimal_density=population_dict['minimal_density'])

    def move(self, entry_id, entry_jd, factor=0.2, in_place=False):

        st_orig = self.get_structure(entry_id)
        st_dest = self.get_structure(entry_jd)

        cm = ClusterMatch(st_orig, st_dest)
        cm.match()

        # pos_orig = np.array(entry_orig['structure']['positions']).reshape((-1, 3))
        # pos_dest = np.array(entry_dest['structure']['positions']).reshape((-1, 3))
        pos_orig = cm.structure1.positions
        pos_dest = cm.structure2.positions

        # Move to a position with negative energy
        reduc = 1
        new_positions = np.array(pos_orig)
        while True:
            new_positions = rotation_move(pos_orig, pos_dest, fraction=reduc * factor)
            new_structure = Structure(positions=new_positions, symbols=st_orig.symbols, periodicity=False)
            lj = LennardJones(new_structure)
            if lj.get_energy() < 0.0:
                break
            reduc -= 0.05
            pcm_log.debug('Effective factor will be reduced to %7.3f, original factor %7.3f' % (reduc * factor, factor))
            if reduc <= 0.0:
                # print 'No movement effective'
                break

        # Avoid condition with atoms too close
        distance_matrix = scipy.spatial.distance_matrix(new_positions, new_positions)
        tmp = np.max(distance_matrix.flatten())
        # print 'Scaling by', tmp
        minimal_distance = np.min((distance_matrix + tmp * np.eye(len(new_positions))).flatten())

        if minimal_distance < 1E-8:
            pcm_log.debug("Null distance between different atoms, no moving")
            new_positions = pos_orig

        if tmp > 5:
            # print 'Big scaling, better not to move'
            new_positions = pos_orig
        else:
            max_cov = np.max(covalent_radius(st_orig.symbols))
            new_positions *= max_cov / minimal_distance

        new_structure = Structure(positions=new_positions, symbols=st_orig.symbols, periodicity=False)
        # print 'Density of cluster', new_structure.density

        if in_place:
            self.unset_properties(entry_id)
            return self.set_structure(entry_id, new_structure)
        else:
            return self.new_entry(new_structure, active=False)

    def evaluate(self, structure, gtol=None):

        if gtol is None:
            gtol = self.target_forces

        positions, forces, energy = lj_compact_evaluate(structure, gtol, self.minimal_density)

        structure.set_positions(positions)
        structure.relocate_to_cm()
        if structure.natom > 2:
            structure.align_inertia_momenta()
        sorted_indices = structure.sort_sites()
        forces = forces[sorted_indices]
        pg = get_point_group(structure, executable='symmol')
        properties = {'forces': generic_serializer(forces), 'energy': energy, 'point_group': pg}
        return structure, properties

    def evaluate_entry(self, entry_id):
        pcm_log.debug('Evaluating %s target density= %7.3F' % (entry_id, self.minimal_density))
        structure = self.get_structure(entry_id)
        structure, properties = self.evaluate(structure, gtol=self.target_forces)
        self.update_properties(entry_id=entry_id, new_properties=properties)
        return self.set_structure(entry_id, structure)

    def refine(self, entry_id, gtol=None):
        if self.refine:
            pcm_log.debug('Evaluating %s target density= %7.3F' % (entry_id, self.minimal_density))
            structure = self.get_structure(entry_id)
            structure, properties = self.evaluate(structure, gtol=gtol)
            self.update_properties(entry_id=entry_id, new_properties=properties)
            return self.set_structure(entry_id, structure)

    def maxforce(self, entry_id):
        return np.max(length_vectors(self.get_forces(entry_id)))

    def refine_progressive(self, entry_id):

        if self.refine:
            inivalue = self.value(entry_id)
            gtol = 10 ** math.ceil(math.log10(self.maxforce(entry_id)))
            while True:
                pcm_log.debug('Local minimization up to %7.3f ' % gtol)
                gtol /= 10
                pcm_log.debug('Evaluating %s target density= %7.3F' % (entry_id, self.minimal_density))
                structure = self.get_structure(entry_id)
                structure, properties = self.evaluate(structure, gtol=gtol)
                if properties['energy'] / structure.natom < inivalue:
                    self.update_properties(entry_id=entry_id, new_properties=properties)
                    return self.set_structure(entry_id, structure)
                else:
                    pcm_log.debug('Relaxation raise value %7.3f < %7.3f' % (inivalue, properties['energy'] /
                                                                            structure.natom))
                if self.maxforce(entry_id) > gtol:
                    pcm_log.debug('I cannot relax more...')
                    break

    def move_random(self, entry_id, factor=0.2, in_place=False, kind='move'):

        entry = self.get_entry(entry_id)
        pos = np.array(entry['structure']['positions']).reshape((-1, 3))
        # Unit Vectors
        uv = unit_vectors(2 * np.random.rand(*pos.shape) - 1)
        new_pos = generic_serializer(pos + factor * uv)

        structure = Structure(positions=new_pos, symbols=entry['structure']['symbols'], periodicity=False)

        if in_place:
            self.update_properties(entry_id=entry_id, new_properties={})
            return self.set_structure(entry_id, structure)
        else:
            structure = Structure(positions=new_pos, symbols=entry['structure']['symbols'], periodicity=False)
            return self.new_entry(structure, active=False)

    def get_structure(self, entry_id):
        entry = self.get_entry(entry_id)
        if 'structure' not in entry:
            raise ValueError('structure is not present on %s' % entry_id)
        if entry['structure'] is None:
            raise ValueError('structure is None for %s' % entry_id)
        return Structure.from_dict(entry['structure'])

    def get_forces(self, entry_id):
        entry = self.get_entry(entry_id, projection={'properties.forces': 1})
        forces = np.array(entry['properties']['forces']).reshape((-1, 3))
        return forces

    def str_entry(self, entry_id):
        structure = self.get_structure(entry_id)
        entry = self.get_entry(entry_id, projection={'properties': 1})
        msg = 'Cluster: LJ%d  Point group: %s  Energy: %7.3f Forces: %7.1E'
        pg = entry['properties']['point_group']
        return msg % (structure.natom, pg, entry['properties']['energy'], self.maxforce(entry_id))

    def new_entry(self, structure, active=True):

        if active and self.direct_evaluation:
            structure, properties = self.evaluate(structure, gtol=self.target_forces)
        else:
            properties = {}
        status = {self.tag: active}
        entry = {'structure': structure.to_dict, 'properties': properties, 'status': status}
        entry_id = self.set_entry(entry)
        pcm_log.debug('Added new entry: %s with tag=%s: %s' % (str(entry_id), self.tag, str(active)))
        return entry_id

    def recover(self):
        data = self.get_population_info()
        if data is not None:
            self.distance_tolerance = data['distance_tolerance']
            self.value_tol = data['value_tol']
            self.name = data['name']
            self.target_forces = data['target_forces']
            self.minimal_density = data['minimal_density']

    def value(self, entry_id):
        entry = self.get_entry(entry_id)
        structure = self.get_structure(entry_id)
        if 'properties' not in entry:
            pcm_log.debug('This entry has no properties %s' % str(entry['_id']))
            return None
        elif entry['properties'] is None:
            return None
        elif 'energy' not in entry['properties']:
            pcm_log.debug('This entry has no energy in properties %s' % str(entry['_id']))
            return None
        else:
            return entry['properties']['energy'] / structure.get_composition().gcd


def rotation_move(pos_orig, pos_dest, fraction):
    new_positions = np.zeros(pos_orig.shape)
    for i in range(len(pos_orig)):
        if length_vector(pos_orig[i]) < 1E-5 or length_vector(pos_dest[i]) < 1E-5:
            new_positions[i] = pos_dest[i]
        else:
            new_positions[i] = rotate_towards_axis(pos_orig[i], pos_dest[i], fraction=fraction)
            uv = new_positions[i] / np.linalg.norm(new_positions[i])
            new_positions[i] = fraction * np.linalg.norm(pos_dest[i]) * uv + (1 - fraction) * \
                                np.linalg.norm(pos_orig[i]) * uv
    return new_positions


def direct_move(pos_orig, pos_dest, fraction):
    return pos_orig + fraction * (pos_dest - pos_orig)


def movement_sweep(pos_orig, pos_dest, symbols, figname='figure.pdf'):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(ncols=1, nrows=3, sharex=True, figsize=(11, 8.5))
    plt.subplots_adjust(left=0.07, bottom=0.07, right=0.98, top=0.98, wspace=0.08, hspace=0.08)

    ee = []
    ff = []
    dd = []
    delta = 2E-3
    xx = np.arange(0.0, 1.0 + 0.9 * delta, delta)

    for f in xx:
        new_positions = direct_move(pos_orig, pos_dest, fraction=f)

        new_structure = Structure(positions=new_positions, symbols=symbols, periodicity=False)
        lj = LennardJones(new_structure)
        ee.append(lj.get_energy())
        ff.append(np.max(lj.get_magnitude_forces()))
        # Distance Matrix
        dm = scipy.spatial.distance_matrix(new_positions, new_positions)
        # Min distance
        md = np.min(np.array(np.array(dm) + 100 * np.eye(len(pos_orig))).flatten())
        dd.append(md)

    ax[0].plot(xx, ee)
    ax[0].set_ylim(min(ee), 0.1)
    ax[1].semilogy(xx, ff)
    ax[2].plot(xx, dd)

    st = Structure(positions=pos_orig, symbols=symbols, periodicity=False)
    lj = LennardJones(st)
    ax[0].plot(0, lj.get_energy(), 'ro')
    ax[1].semilogy(0, np.max(lj.get_magnitude_forces()), 'ro')
    dm = scipy.spatial.distance_matrix(lj.structure.positions, lj.structure.positions)
    md = np.min(np.array(np.array(dm) + 100 * np.eye(len(pos_orig))).flatten())
    ax[2].plot(0, md, 'ro')

    st = Structure(positions=pos_dest, symbols=symbols, periodicity=False)
    lj = LennardJones(st)
    ax[0].plot(1, lj.get_energy(), 'ro')

    ax[1].semilogy(1, np.max(lj.get_magnitude_forces()), 'ro')
    dm = scipy.spatial.distance_matrix(lj.structure.positions, lj.structure.positions)
    md = np.min(np.array(np.array(dm) + 100 * np.eye(len(pos_orig)).flatten()))
    ax[2].plot(1, md, 'ro')

    ax[2].set_xlim(-0.01, 1.01)

    ax[0].set_ylabel('Energy')
    ax[1].set_ylabel('Max Force')
    ax[2].set_ylabel('Minimal inter atomic distance')

    plt.savefig(figname)



#!/usr/bin/env python

import os                              # Operating System module (create directories)
import time                            # To create a sleeping timer
import random
import subprocess                      # To launch an external command (Such as maise)
from multiprocessing import Process    # Multiprocessing python (Concurrency)


def maise_worker(case, worker_args):
    # os.chdir(workdir)
    # subprocess.call(['maise'])
    random.seed()
    workdir = "MAISE_%03d" % case
    if not os.path.isdir(workdir):
        os.mkdir(workdir)
    wf = open(workdir+'/EXECTUING MAISE', 'w')
    wf.write('Put results here')
    ttl = random.randint(10, 40)
    print('Evaluating case: %d this calculation will take %d seconds' % (case, ttl))
    time.sleep(ttl)
    wf.close()


def run(to_evaluate, worker, nconcurrent=1, sleeping_time=120):
    """
    For a given list of directories to evaluate, start processing concurrently the worker until all directories are
    processed

    :param worker:        Function to evaluate for each case
    :param sleeping_time: Number of seconds to wait until checking for complete jobs
    :param to_evaluate: (list) List of directories to evaluate
    :param nconcurrent: (int)  number of concurrent evaluations of the worker
    """

    procs = []
    worker_args = (None,)  # This is a dummy argument, not playing any role here, it must be a tuple always

    # Creating a list to store the 'nconcurrent' running jobs
    for i in range(nconcurrent):
        procs.append(None)

    # Main loop looking permanently for candidates for evaluation
    while True:

        index = 0
        currently_evaluating = 0
        for j in range(nconcurrent):
            if procs[j] is not None and procs[j].is_alive():
                currently_evaluating += 1
        print('Candidates to evaluate: %d  Candidates in evaluation: %d' % (len(to_evaluate), currently_evaluating))

        while index < len(to_evaluate):

            slot = None
            while True:
                for j in range(nconcurrent):
                    if procs[j] is None or not procs[j].is_alive():
                        slot = j
                        break
                if slot is None:
                    time.sleep(sleeping_time)
                else:
                    break

            case = to_evaluate[index]
            print("I will run case: %d with index: %d on slot %d/%d" % (case, index, slot+1, nconcurrent))

            # This is the actual call to the worker, it must be a function with 4 arguments:
            # The database settings, the entry identifier, the working directory and arguments for the worker
            procs[slot] = Process(target=worker, args=(case, worker_args))
            procs[slot].start()
            time.sleep(1)

            index += 1
        time.sleep(sleeping_time)

        # Condition to finish execution, if removed the code will run forever looking for more cases to run
        if index >= len(to_evaluate):
            break


if __name__ == '__main__':

    to_evaluate = range(100)  # Creating 100 cases to evaluate
    nconcurrent = 16          # Number of concurrent executions

    run(to_evaluate=to_evaluate, nconcurrent=nconcurrent, worker=maise_worker, sleeping_time=5)





"""
Delaunay Reduction
"""

import numpy as _np


def get_reduced_bases(cell, tolerance=1e-5):
    """
    This is an implementation of Delaunay reduction.
    Some information is found in International table.

    :param cell: (numpy.ndarray) Cell parameters dimension 3x3
    :param tolerance: (float) Tolerance used to reduce basis
    :return: (numpy.ndarray) a new cell
    """
    return get_delaunay_reduction(cell, tolerance)


def get_delaunay_reduction(lattice, tolerance):
    """
    Return a reduced cell using the delanuay reduction

    :param lattice: (numpy.ndarray) the cell parameters
    :param tolerance: (float) Tolerance used to reduce basis
    :return: (numpy.ndarray) a new cell
    """
    extended_bases = _np.zeros((4, 3), dtype=float)
    extended_bases[:3, :] = lattice
    extended_bases[3] = -_np.sum(lattice, axis=0)

    i = 0
    for i in range(100):
        if reduce_bases(extended_bases, tolerance):
            break
    if i == 99:
        print("Delaunary reduction is failed.")

    shortest = get_shortest_bases_from_extented_bases(extended_bases, tolerance)

    return shortest


def reduce_bases(extended_bases, tolerance):
    """
    Reduces the basis with a given tolerance

    :param extended_bases:
    :param tolerance: (float) Tolerance used by reduction algorithm
    """
    metric = _np.dot(extended_bases, extended_bases.transpose())
    for i in range(4):
        for j in range(i + 1, 4):
            if metric[i][j] > tolerance:
                for k in range(4):
                    if (not k == i) and (not k == j):
                        extended_bases[k] += extended_bases[i]
                extended_bases[i] = -extended_bases[i]
                extended_bases[j] = extended_bases[j]
                return False

    # Reduction is completed.
    # All non diagonal elements of metric tensor is negative.
    return True


def get_shortest_bases_from_extented_bases(extended_bases, tolerance):
    # print 'get_shortest_bases_from_extented_bases',tolerance

    def mycmp(x, y):
        return cmp(_np.vdot(x, x), _np.vdot(y, y))

    basis = _np.zeros((7, 3), dtype=float)
    basis[:4] = extended_bases
    basis[4] = extended_bases[0] + extended_bases[1]
    basis[5] = extended_bases[1] + extended_bases[2]
    basis[6] = extended_bases[2] + extended_bases[0]
    # Sort bases by the lengthes (shorter is earlier)
    basis = sorted(basis, cmp=mycmp)

    # Choose shortest and linearly independent three bases
    # This algorithm may not be perfect.
    for i in range(7):
        for j in range(i + 1, 7):
            for k in range(j + 1, 7):
                if abs(_np.linalg.det([basis[i], basis[j], basis[k]])) > tolerance:
                    return _np.array([basis[i], basis[j], basis[k]])

    print("Delaunary reduction is failed.")
    return basis[:3]



from .searcher import Searcher
from pychemia import pcm_log


class ParticleSwarm(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):
        """
        Implementation fo the Particle Swarm Optimization algorithm for global minimization

        :param population:
        :param params: (dict) Parameters to setup the Searcher. For PSO the parameters are:
            'beta0': A value in the range (0,1) that defines how close the mobile structure will be from to
             the target structure
            'alpha0': Factor of scale for the random movement
            'delta': How the random change decreases with time
        :param generation_size: (int)
        :param stabilization_limit: (int)
        :return:
        """
        # Initializing objects
        Searcher.__init__(self, population, generation_size, stabilization_limit)
        # Parameters
        self.beta0 = None
        self.alpha0 = None
        self.delta = None
        self.set_params(params)

    def set_params(self, params):
        # Default parameters
        self.beta0 = 0.9
        self.alpha0 = 0.3
        self.delta = 0.9
        if params is None:
            params = {}
        if 'beta0' in params:
            self.beta0 = params['beta0']
        if 'alpha0' in params:
            self.alpha0 = params['alpha0']
        if 'delta' in params:
            self.delta = params['delta']

    def get_params(self):
        return {'beta0': self.beta0,
                'alpha0': self.alpha0,
                'delta': self.delta}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.actives_in_generation)
        pcm_log.info('Size of selection : %d' % len(selection))

        # Minus sign because we are searching for minima
        intensity = self.population.get_values(selection)
        for entry_id in intensity:
            intensity[entry_id] *= -1

        moves = {}
        new_selection = {}
        for entry_id in selection:
            new_selection[entry_id] = None
            moves[entry_id] = 0

        # Move all the particles (Except the elite)
        # as the selection is sorted it means that the first one will no move
        pcm_log.debug('No Moving %d %s. Intensity: %7.3f' % (0, str(selection[0]), intensity[selection[0]]))
        for i in range(1, len(selection)):
            entry_id = selection[i]
            pcm_log.debug('Moving %d %s. Intensity: %7.3f' % (i, str(entry_id), intensity[entry_id]))

            new_selection[entry_id] = self.population.move(entry_id, selection[0], factor=self.beta0, in_place=False)
            factor = self.alpha0 * (self.delta ** self.current_generation)
            self.population.move_random(new_selection[entry_id], factor=factor, in_place=True)

        for entry_id in selection:
            if new_selection[entry_id] is not None:
                pcm_log.debug('[%s] Moved to: %s (%d moves)' %
                              (str(entry_id), new_selection[entry_id], moves[entry_id]))
                self.replace_by_other(entry_id, new_selection[entry_id],
                                      reason='Moved %d times' % moves[entry_id])
                self.population.enable(new_selection[entry_id])
            else:
                pcm_log.debug('[%s] Promoted to new generation' % str(entry_id))
                self.pass_to_new_generation(entry_id, reason='The best')



import os
import re
import subprocess
import numpy as np
import numbers
from ..codes import CodeRun
from pychemia.crystal import KPoints
from pychemia import Structure, pcm_log
from pychemia.utils.serializer import generic_serializer


class DFTBplus(CodeRun):
    def __init__(self, workdir='.'):

        if not os.path.lexists(workdir):
            os.mkdir(workdir)

        CodeRun.__init__(self, executable='dftb+', workdir=workdir)
        self.geometry = {}
        self.driver = {}
        self.hamiltonian = {}
        self.options = {}
        self.analysis = {}
        self.parser_options = {}
        self.slater_koster = []
        self.structure = Structure()
        self.runner = None
        self.kpoints = None
        self.stdout_file = None
        self.output = None
        self.kp_density = None

    def initialize(self, structure, kpoints=None, kp_density=10000):
        assert structure.is_crystal
        assert structure.is_perfect
        self.structure = structure
        self.get_geometry()
        self.kp_density = kp_density
        if kpoints is None:
            kpoints = KPoints.optimized_grid(self.structure.lattice, kp_density=self.kp_density, force_odd=True)
        self.kpoints = kpoints

    def set_inputs(self):
        self.write_input(filename=self.workdir + os.sep + 'dftb_in.hsd')
        self.link_slater_koster()

    def get_outputs(self):
        self.output = read_detailed_out(filename=self.workdir + os.sep + 'detailed.out')

    def run_status(self):
        if self.runner is None:
            pcm_log.info('DFTB+ not finish')
            filename = self.workdir + os.sep + 'dftb_stdout.log'
            if os.path.exists(filename):
                booleans, geom_optimization, stats = read_dftb_stdout(filename=filename)
                pcm_log.debug(str(booleans))
                pcm_log.debug(str(geom_optimization))
                pcm_log.debug(str(stats))
            return
        if self.runner.poll() == 0:
            pcm_log.info('DFTB+ complete normally')

    def finalize(self):
        self.stdout_file.close()

    @staticmethod
    def read_input(filepath='dftb_in.hsd'):
        """
        Read an Input for DFTB+
        Support a very restricted subset of
        Human-readable Structured Data (HSD)

        :param filepath: (str) Filename to read as DFTB+ input
        :return: (dict)
        """
        rf = open(filepath, 'r')
        ret = []
        gen_format = False
        value = None
        level = 0
        current_container = ret
        tree_pos = [ret]
        for line in rf.readlines():
            line = line.partition('#')[0]
            line = line.rstrip()
            print('line--->', line)
            print('return->', ret)
            # print 'container',current_container
            # print 'tree_pos',tree_pos
            print(80 * '-')
            if line.strip() == '':
                continue
            elif '=' in line and line.strip()[-1] == '{':
                leftside = line.split('=')[0].strip()
                rightside = line.split('=')[1].split()[0].strip()
                ret.append({'_left': leftside, '_rigth': rightside})
                current_container = ret[-1]
                tree_pos.append(current_container)
                if rightside == 'GenFormat':
                    gen_format = True
                level += 1
                print('container', current_container)
            elif gen_format and line.strip() != '}':
                if value is None:
                    value = line
                else:
                    value += line
            elif line.strip() == '}':
                if gen_format:
                    current_container['value'] = value
                    value = None
                    gen_format = False
                tree_pos = tree_pos[:-1]
                current_container = tree_pos[-1]
                level -= 1
            elif '=' in line and level > 0:
                current_container[line.split('=')[0].strip()] = line.split('=')[1]
            elif ('=' not in line) and (line.strip()[-1] == '{'):
                if level > 0:
                    current_container[line.split()[0].strip()] = {}
                    current_container = current_container[line.split()[0].strip()]
                    tree_pos.append(current_container)
                    level += 1
                else:
                    ret.append({line.strip().split()[0]: {}})
                    current_container = ret[-1]
                    tree_pos.append(current_container)
                    level += 1
            if line.strip()[-2:] == '{}':
                if level == 0:
                    ret.append({line.split()[0]: '{}'})
                else:
                    raise ValueError('Line not parsed correctly: ' + line)
            else:
                raise ValueError('Line not parsed correctly: ' + line)
        return ret

    def get_geometry(self):
        assert (self.structure is not None)

        self.geometry['TypeNames'] = list(self.structure.species)
        self.geometry['TypesAndCoordinates'] = {'units': 'Angstrom'}
        indices = []
        for isite in self.structure:
            indices.append(self.geometry['TypeNames'].index(isite.symbols[0]) + 1)
        self.geometry['TypesAndCoordinates']['types'] = np.array(indices)
        self.geometry['TypesAndCoordinates']['positions'] = self.structure.positions
        self.geometry['Periodic'] = True
        self.geometry['LatticeVectors'] = {'units': 'Angstrom', 'coordinates': self.structure.cell}

    def basic_driver(self, lattice_optimization=True, maxforce=1E-2, maxsteps=50):

        self.driver['method'] = 'ConjugateGradient'
        self.driver['MaxForceComponent'] = maxforce
        self.driver['MaxSteps'] = maxsteps
        self.driver['LatticeOpt'] = lattice_optimization
        self.driver['MovedAtoms'] = '1:-1'

    def set_static(self):
        self.driver = {}

    def basic_options(self):
        self.options = {'WriteDetailedXML': True}

    def basic_analysis(self):
        self.analysis = {'ProjectStates': []}

    def basic_parser_options(self):
        self.parser_options = {'WriteXMLInput': True}

    @staticmethod
    def print_block(name, block):

        if block == {}:
            return '\n' + name + ' = {}\n'

        ret = '\n' + name + ' = '
        if 'method' in block:
            ret += block['method'] + ' {\n'
        else:
            ret += ' {\n'
        for ivar in sorted([x for x in block if x != 'method']):
            if isinstance(block[ivar], str):
                ret += ' ' + ivar + ' = ' + DFTBplus._write_string(block[ivar])
            elif isinstance(block[ivar], bool):
                ret += ' ' + ivar + ' = ' + DFTBplus._write_bool(block[ivar])
            elif isinstance(block[ivar], numbers.Number):
                ret += ' ' + ivar + ' = ' + DFTBplus._write_number(block[ivar])
            elif isinstance(block[ivar], dict):
                ret += ' ' + ivar + DFTBplus._write_dict(block[ivar])
            elif isinstance(block[ivar], list):
                ret += ' ' + ivar + ' = ' + DFTBplus._write_list(block[ivar])
        ret += '}\n'
        return ret

    def basic_hamiltonian(self, scc=True, slater='auto'):
        self.hamiltonian['method'] = 'DFTB'
        self.hamiltonian['SCC'] = scc
        if slater == 'auto':
            self.hamiltonian['SlaterKosterFiles'] = {}
            for i in self.structure.species:
                for j in self.structure.species:
                    self.hamiltonian['SlaterKosterFiles'][i + '-' + j] = '"' + i + '-' + j + '.skf' + '"'

        self.hamiltonian['MaxAngularMomentum'] = {}
        max_angular_momentum = self.get_max_angular_momentum()
        for i in max_angular_momentum:
            self.hamiltonian['MaxAngularMomentum'][i] = '"' + max_angular_momentum[i] + '"'

        self.hamiltonian['KPointsAndWeights'] = self.set_kpoints()

    @staticmethod
    def _write_dict(the_dict):
        ret = ' '
        if 'name' in the_dict:
            ret += ' = '
            ret += the_dict['name'] + ' {'
            if 'value' in the_dict and isinstance(the_dict['value'], np.ndarray):
                ret += DFTBplus._write_ndarray(the_dict['value'])
        elif 'units' in the_dict:
            ret += '[' + the_dict['units'] + '] = {'
            if 'coordinates' in the_dict:
                if isinstance(the_dict['coordinates'], np.ndarray):
                    ret += DFTBplus._write_ndarray(the_dict['coordinates'])
            elif 'types' in the_dict:
                ret += '\n'
                for i in range(len(the_dict['types'])):
                    ret += " %3d %10.7f %10.7f %10.7f\n" % (the_dict['types'][i],
                                                            the_dict['positions'][i][0],
                                                            the_dict['positions'][i][1],
                                                            the_dict['positions'][i][2])
        else:
            ret += ' = {\n'
            for ivar in the_dict:
                ret += ' ' + ivar + ' = ' + the_dict[ivar] + '\n'
        ret += ' }\n'
        return ret

    @staticmethod
    def _write_ndarray(the_ndarray):
        ret = ''
        if the_ndarray.ndim == 1:
            ret += '\n'
            for i in the_ndarray:
                ret += ' %10.7f ' % i
        elif the_ndarray.ndim == 2:
            ret += '\n'
            for i in the_ndarray:
                for j in i:
                    ret += ' %10.7f ' % j
                ret += '\n'
        return ret

    @staticmethod
    def _write_bool(value):
        if value:
            return ' Yes\n'
        else:
            return ' No\n'

    @staticmethod
    def _write_number(value):
        return str(value) + '\n'

    @staticmethod
    def _write_list(value):
        ret = ' {'
        for i in value:
            ret += ' ' + i
        ret += ' }\n'
        return ret

    @staticmethod
    def _write_string(value):
        return value + '\n'

    def basic_input(self):
        self.basic_driver()
        if not self.slater_koster:
            pcm_log.debug('The Slater-Koster files were not selected')
            pcm_log.debug('The Hamiltonian could not be setup')
        else:
            self.basic_hamiltonian()
        self.basic_options()
        self.basic_analysis()
        self.basic_parser_options()

    def print_all(self):
        ret = self.print_block('Geometry', self.geometry)
        ret += self.print_block('Driver', self.driver)
        ret += self.print_block('Hamiltonian', self.hamiltonian)
        ret += self.print_block('Options', self.options)
        ret += self.print_block('Analysis', self.analysis)
        ret += self.print_block('ParserOptions', self.parser_options)
        return ret

    def write_input(self, filename='dftb_in.hsd'):

        wf = open(filename, 'w')
        wf.write(self.print_all())
        wf.close()

    def set_slater_koster(self, search_paths):
        ret = True
        if isinstance(search_paths, str):
            search_paths = [search_paths]
        elif not isinstance(search_paths, list):
            raise ValueError('search_path is not an string or list')
        self.slater_koster = []
        for ispecie in self.structure.species:
            for jspecie in self.structure.species:
                pair = ispecie + '-' + jspecie
                pair_found = False
                path = self.workdir + os.sep + pair + '.skf'
                if os.path.exists(path):
                    self.slater_koster.append(path)
                    pair_found = True
                else:
                    for ipath in search_paths:
                        path = ipath + os.sep + pair + '.skf'
                        if os.path.exists(ipath + os.sep + ispecie + '-' + jspecie + '.skf'):
                            self.slater_koster.append(path)
                            pair_found = True
                if pair_found:
                    # print 'Slater-Koster %7s : %s' % (pair, path)
                    pcm_log.debug('Slater-Koster %7s : %s' % (pair, path))
                else:
                    pcm_log.debug('ERROR: Slater_Koster for ' + pair + ' not found')
                    ret = False
        return ret

    def link_slater_koster(self):
        """
        Create symbolic links to Slater-Koster files not already on the workdir
        """
        for i in self.slater_koster:
            filename = self.workdir + os.sep + os.path.basename(i)
            if os.path.exists(filename):
                pass
            elif os.path.lexists(filename):
                os.remove(filename)
                os.symlink(os.path.abspath(i), filename)
            else:
                os.symlink(os.path.abspath(i), filename)

    @staticmethod
    def get_shells(slater_koster):

        ret = {}
        rf = open(slater_koster, 'r')
        xmldata = ''
        isxml = False
        for iline in rf.readlines():
            if iline.strip() == '<Documentation>':
                isxml = True
            elif iline.strip() == '</Documentation>':
                isxml = False
                xmldata += iline
            if isxml:
                xmldata += iline

        import xml.etree.ElementTree
        root = xml.etree.ElementTree.fromstring(xmldata)
        specie1 = root.findall('./General/Element1')[0].text
        specie2 = None
        if root.findall('./General/Element2'):
            specie2 = root.findall('./General/Element2')[0].text

        for ibasis in root.findall('./SK_table/Basis'):
            if ibasis.attrib['atom'] == '1':
                specie = specie1
            elif ibasis.attrib['atom'] == '2':
                specie = specie2
            else:
                specie = None
            ret[specie] = ibasis.find('./Shells').text.strip()
        return ret

    def get_all_shells(self):
        ret = {}
        for islater in self.slater_koster:
            shells = self.get_shells(islater)
            for jspecie in shells:
                if jspecie in ret:
                    if not shells[jspecie] in ret[jspecie]:
                        ret[jspecie].append(shells[jspecie])
                else:
                    ret[jspecie] = [shells[jspecie]]
        return ret

    def get_max_angular_momentum(self):

        ret = {}
        maxang = None
        shells = self.get_all_shells()
        for ispecie in shells:
            assert (len(shells[ispecie]) == 1)
            if ispecie in self.structure.species:
                for ishell in shells[ispecie]:
                    # Take the last orbital from a shell like '2s 2p' => p
                    maxang = ishell.split()[-1][-1]
                ret[ispecie] = maxang
        return ret

    def set_kpoints(self):
        assert isinstance(self.kpoints, KPoints)
        ret = None
        if self.kpoints.kmode in ['cartesian', 'reciprocal']:
            ret = np.zeros((self.kpoints.nkpt, 4))
            ret[0:3, 0:3] = self.kpoints.kpoints_list
            ret[:, 3] = self.kpoints.weights
        elif self.kpoints.kmode in ['gamma', 'monkhorst-pack']:
            ret = {'name': 'SupercellFolding', 'value': np.zeros((4, 3))}
            ret['value'][:3, :3] = np.diag(self.kpoints.grid)
            if self.kpoints.kmode == 'gamma':
                ret['value'][3, :] = np.zeros(3)
            elif self.kpoints.kmode == 'monkhorst-pack':
                for i in range(3):
                    # Test if grid is odd
                    if self.kpoints.grid[i] & 1:
                        # ODD
                        ret['value'][3, i] = 0
                    else:
                        # EVEN
                        ret['value'][3, i] = 0.5
        return ret

    def roll_outputs(self, value):
        for ifile in ['detailed.out', 'detailed.xml', 'dftb_stdout.log', 'geo_end.gen', 'geo_end.xyz', 'dftb_in.hsd']:
            if os.path.exists(self.workdir + os.sep + ifile):
                os.rename(self.workdir + os.sep + ifile, self.workdir + os.sep + ('%03d_%s' % (value, ifile)))


def read_geometry_gen(filename):
    rf = open(filename, 'r')
    line = rf.readline()
    natom = int(line.split()[0])
    kind = line.split()[1].strip()
    if kind in ['S', 'F']:
        periodic = True
    elif kind == 'C':
        periodic = False
    else:
        raise ValueError('Wrong type of geometry' + kind + ". Should be ('F', 'S' or 'C')")
    species = rf.readline().split()
    coords = np.zeros((natom, 3))
    symbols = []
    for i in range(natom):
        line = rf.readline()
        coords[i, :] = np.array([float(x) for x in line.split()[2:]])
        # The indices for atoms on a gen file are in the second col [1] and start in 1
        symbols.append(species[int(line.split()[1]) - 1])
    if periodic:
        line = rf.readline()
        coords_origin = np.array(line.split())
        cell = np.zeros((3, 3))
        for i in range(3):
            line = rf.readline()
            cell[i, :] = np.array([float(x) for x in line.split()])

    if kind == 'F':
        return Structure(symbols=symbols, cell=cell, reduced=coords, periodicity=True)
    elif kind == 'S':
        return Structure(symbols=symbols, cell=cell, positions=coords, periodicity=True)
    elif kind == 'C':
        return Structure(symbols=symbols, positions=coords, periodicity=False)


def read_detailed_out(filename='detailed.out'):
    """
    Read a 'detailed.out' file and extract the values
    of forces, stress and total energy
    If those any of those three properties are not present
    its value will be set to None

    :param filename: The filename in the format of DFTB+ 'detailed.out'
    :return: tuple (forces, stress, total_energy)
    """
    ret = {}

    rf = open(filename, 'r')
    data = rf.read()

    # Extracting the data from 'detailed.out'
    forces = re.findall(r'Total\s*Forces\s*([.\-+E\d\s]+)\n', data)
    stress = re.findall(r'Total\s*stress\s*tensor\s*([.\-+E\d\s]+)\n', data)
    total_energy = re.findall(r'Total\s*energy:\s*[.\-+E\d\s]*\s*H\s*([.\-\d\sE]+)\s*eV', data)

    max_force = re.findall(r'Max force for moved atoms::\s*([.\-+E\d\s]+)\s*au', data)
    if len(max_force) > 0:
        max_force = float(max_force[-1])
    else:
        max_force = None

    max_deriv = re.findall(r'Maximal derivative component:\s*([.\-+E\d\s]+)\s*au', data)
    if len(max_deriv) > 0:
        max_deriv = float(max_deriv[-1])
    else:
        max_deriv = None

    if forces:
        forces = np.array(forces[0].split(), dtype=float).reshape((-1, 3))
        # pcm_log.debug('Forces :\n'+str(forces))
    else:
        forces = None

    if stress:
        stress = np.array(stress[0].split(), dtype=float).reshape((3, 3))
        # pcm_log.debug('Stress :\n' + str(stress))
    else:
        stress = None

    if total_energy:
        total_energy = float(total_energy[0])
        # pcm_log.debug('Energy :' + str(total_energy))
    else:
        total_energy = None

    scc = re.findall('iSCC Total electronic\s+Diff electronic\s+SCC error\s+([\s\dE+-.]*)', data)
    if len(scc) > 0:
        ret['SCC'] = {}
        ret['SCC']['iSCC'] = int(scc[0].split()[0])
        ret['SCC']['Total_electronic'] = float(scc[0].split()[1])
        ret['SCC']['Diff_electronic'] = float(scc[0].split()[2])
        ret['SCC']['SCC_error'] = float(scc[0].split()[3])

    ret['forces'] = forces
    ret['stress'] = stress
    ret['total_energy'] = total_energy
    if max_force is not None:
        ret['max_force'] = max_force
    if max_deriv is not None:
        ret['max_deriv'] = max_deriv
    return ret


def read_dftb_stdout(filename='dftb_stdout.log'):
    """
    Read the standard output stored in a file
    (by default 'dftb_stdout.log') and extract
    relevant information useful for a relaxation
    procedure

    :param filename: The standard output stored in a file
    :return:
    """

    rf = open(filename, 'r')
    data = rf.read()

    ret = {}

    start_ini = re.findall(r'Starting initialization...[\s-]+([.\s\d\w\-:(),+]+)---', data)
    if len(start_ini) == 1:
        start_ini = start_ini[0].replace('\n  ', '  ').split('\n')
        ret['Starting_Initialization'] = {}
        for line in start_ini:
            if len(line.split(':')) == 2:
                left, right = line.split(':')
            elif line[:3] != '---':
                left = line[:29].replace(':', '')
                right = line[29:]
            ret['Starting_Initialization'][left.strip()] = parse_value(right.strip())

            if left.strip() == 'K-points and weights':
                right = right.replace(':', '')
                right = np.array(right.split(), dtype=float).reshape((-1, 5))
                ret['Starting_Initialization'][left.strip()] = {'kpoints': generic_serializer(right[:, 1:-1]),
                                                                'weights': generic_serializer(right[:, -1])}

    geom_blocks = re.findall(r'Geometry step:([\^*.\s\d\w\-:(),+]+)Pa', data)

    ret['Geometry_Steps'] = []
    for iblock in geom_blocks:
        ib = iblock.split('\n')
        if 'Lattice' in ib[0]:
            tmp = {'iter': int(ib[0].split(',')[0]), 'scc': {}}
        else:
            tmp = {'iter': int(ib[0]), 'scc': {}}
        tmp['scc']['tot_electronic'] = []
        tmp['scc']['diff_electronic'] = []
        tmp['scc']['scc_error'] = []
        for iline in ib[4:]:
            fields = iline.split()
            if len(fields) > 0 and fields[0].strip().isdigit():
                tmp['scc']['tot_electronic'].append(float(fields[1]))
                tmp['scc']['diff_electronic'].append(float(fields[2]))
                tmp['scc']['scc_error'].append(float(fields[3]))
            elif len(iline.split(':')) == 2:
                left, right = iline.split(':')
                tmp[left.strip()] = {'value': float(right.split()[0]), 'units': right.split()[1]}

        ret['Geometry_Steps'].append(tmp)

    max_force = re.findall(r'Maximal force component:\s*([\s\dE+-.]+)\s*\n', data)
    if len(max_force) > 0:
        ret['max_force'] = float(max_force[-1])

    if re.findall('Geometry did NOT converge', data):
        pcm_log.debug('Convergence not achieved!')
        ret['ion_convergence'] = False
    else:
        ret['ion_convergence'] = True

    if re.findall('Geometry converged', data):
        pcm_log.debug('Convergence achieved!')
        ret['ion_convergence'] = True
    else:
        ret['ion_convergence'] = False
    return ret


def parse_value(value):
    try:
        ret = int(value)
    except ValueError:
        try:
            ret = float(value)
        except ValueError:
            if value == 'Yes':
                ret = True
            elif value == 'No':
                ret = False
            else:
                ret = value
    return ret



from abc import ABCMeta, abstractmethod
import os
import subprocess
import collections
import psutil
import shlex
from pychemia import pcm_log

class CodeRun:
    __metaclass__ = ABCMeta

    def __init__(self, executable, workdir='.', use_mpi=False):
        """
        CodeRun is the superclass defining the operations for running codes either directly or asynchronously via a
        submission script on a cluster.

        :param executable:   Name or Path to the executable if can be found on the $PATH or path to the executable
        :param workdir:  Path to a folder where input and output will be located (Default: '.')
        :param use_mpi:  True if code relies on MPI for execution (Default: False)

        """
        self.stdin_file = None
        self.stdout_file = None
        self.stderr_file = None
        self.stdin_filename = None
        self.stdout_filename = None
        self.stderr_filename = None
        self.input_path = None
        self.input = None
        self.executable = executable
        self.workdir = workdir
        self.use_mpi = use_mpi
        self.runner = None

    @abstractmethod
    def set_inputs(self):
        """
        This method must be implemented by child classes, it should write all the input files and prepare environment
        for execution

        :return: None
        """
        pass

    @abstractmethod
    def get_outputs(self):
        """
        This method must be implemented by child classes, it check the existance of output files and prepare the
        reading and parsing of output data.

        :return: None
        """
        pass

    def run(self, num_threads=None, mpi_num_procs=None, nodefile=None, wait=True, verbose=False):
        """
        Run the executable and return a reference to the subprocess
        created. The execution can set a number of threading variables to control their number.
        If the code uses MPI the variable mpi_num_procs control the number of processes used.

        :param num_threads: Control the number of threads in multithreading applications. There are several environment
                            variables used for that purpose. The default is None, meaning that not variable is set,
                            if an integer is given only OMP_NUM_THREADS is set dynamically for the run if a dictionary
                            of variables is provided with keys OMP_NUM_THREADS, OPENBLAS_NUM_THREADS, GOTO_NUM_THREADS
                            and MKL_NUM_THREADS, the corresponding integers associated to those variables are used to
                            set the number of corresponding threads.

        :param mpi_num_procs: Number of MPI processes, if None and the code uses MPI, the default is to set the maximum
                              number of cores available on the system.

        :param nodefile: If a nodefile is provided, and the code uses MPI, the contents are used to select the number of
                         of processes created. Setting nodefile does not override the use of mpi_num_procs but if
                         mpi_num_procs is None it will assume the number of processes equal to the machines
                         declared on nodefile.

        :param wait: If True the process waits until the command is completed, otherwise a Popen instance is returned

        :param verbose: Print extra information before and after the execution

        :return:
        """
        cwd = os.getcwd()
        pcm_log.debug("Current location=%s, workdir=%s" % (cwd, self.workdir))
        os.chdir(self.workdir)

        env_vars = ''
        if num_threads is not None:
            if isinstance(num_threads, int):
                env_vars = 'OMP_NUM_THREADS='+str(num_threads)
        elif isinstance(num_threads, dict):
            for evar in ['OMP_NUM_THREADS', 'OPENBLAS_NUM_THREADS', 'GOTO_NUM_THREADS', 'MKL_NUM_THREADS']:
                if evar in num_threads and isinstance(num_threads[evar], int) and num_threads[evar] > 0:
                    env_vars += evar+'='+str(num_threads[evar])+' '

                subprocess.check_output('export %s=%d' % (evar, num_threads), shell=True)
                envvars=subprocess.check_output('echo $%s' % evar, shell=True)
                print(envvars.decode())

        if self.stdin_filename is not None:
            self.stdin_file = open(self.stdin_filename, 'r')
        if self.stdout_filename is not None:
            self.stdout_file = open(self.stdout_filename, 'w')
        if self.stderr_filename is not None:
            self.stderr_file = open(self.stderr_filename, 'w')

        # Checking availability of executable
        if not os.path.exists(self.executable):

            try:
                which_bin = subprocess.check_output('which %s' % self.executable, shell=True)
            except subprocess.CalledProcessError:
                raise ValueError('ERROR: Executable %s could not be found as an absolute, relative or via the $PATH variable' %
                      self.executable)
                
            exec_path = which_bin.decode('utf8').strip()
        else:
            exec_path = self.executable

        pcm_log.debug("Executable: %s " % exec_path)

        if self.use_mpi:

            np = 2
            if mpi_num_procs is None:
                np = psutil.cpu_count(logical=False)
            elif isinstance(mpi_num_procs, int):
                np = mpi_num_procs
            else:
                print("WARNING: Declared variable mpi_num_procs is not integer, defaulting to 2 process")

            command = 'mpirun -n %d %s' % (np, self.executable)
        else:
            command = "%s" % self.executable

        pcm_log.debug("Running: %s" % command)
        process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=self.stdin_file)
        while True:
            output = process.stdout.readline()
            if output == b'' and process.poll() is not None:
                break
            if output != b'' and process.poll() is not None:
                pcm_log.debug("process.poll() is %s" % process.poll())
                pcm_log.debug(output)
            if output and verbose:
                print(output.decode(), end='')
        rc = process.poll()
        self.runner = process
        os.chdir(cwd)
        return process

#       if wait:
#            self.runner.wait()
#            if verbose:
#                print("Program finished with returncode: %d" % self.runner.returncode)
#        os.chdir(cwd)
#        return self.runner


class CodeInput(collections.abc.MutableMapping):

    __metaclass__ = ABCMeta

    def __init__(self):
        self.input_file = None
        self.variables = {}

    def __contains__(self, x):
        if not self.is_hierarchical:
            return x in self.variables
        else:
            return x[1] in self.variables[x[0]]

    def __delitem__(self, x):
        if not self.is_hierarchical:
            return self.variables.__delitem__(x)
        else:
            return self.variables[x[0]].__delitem__(x[1])

    def __setitem__(self, x, value):
        if not self.is_hierarchical:
            return self.variables.__setitem__(x, value)
        else:
            return self.variables[x[0]].__setitem__(x[1], value)

    def __getitem__(self, x):
        if not self.is_hierarchical:
            return self.variables.__getitem__(x)
        else:
            return self.variables[x[0]].__getitem__(x[1])

    def __iter__(self):
        return self.variables.__iter__()

    def __len__(self):
        return self.variables.__len__()

    @abstractmethod
    def read(self):
        pass

    @abstractmethod
    def __str__(self):
        pass

    def write(self, filename=None):
        """
        Write an input object into a text
        file that ABINIT can use as an input
        file

        Args:
            filename:
                The 'abinit.in' filename that will be written
        """
        if filename is None:
            if self.input_file is not None:
                filename = self.input_file
            else:
                raise ValueError("Not filename indicated")
        wf = open(filename, 'w')
        wf.write(self.__str__())
        wf.close()

    def has_variable(self, varname, section=None):
        if self.is_hierarchical:
            if section is None:
                raise ValueError('ERROR: Input variables are hierachical and not section was declared')
            else:
                if section in self.variables and varname in self.variables[section]:
                    return True
                else:
                    return False
        else:
            if varname in self.variables:
                return True
            else:
                return False

    def get_variable(self, varname, section=None):

        if self.is_hierarchical:
            if section is None:
                raise ValueError('ERROR: Input variables are hierachical and not section was declared')
            else:
                if self.has_variable(varname, section=section):
                    return self.variables[section][varname]
                else:
                    return None
        else:
            if self.has_variable(varname):
                return self.variables[varname]
            else:
                return None

    def set_variable(self, varname, value, section=None):
        if self.is_hierarchical:
            if section is None:
                raise ValueError('ERROR: Input variables are hierarchical and not section was declared')
            else:
                self.variables[section][varname] = value
        else:
            self.variables[varname] = value

    @property
    def get_number_variables(self):
        if not self.is_hierarchical:
            return len(self.variables)
        else:
            ret = {}
            for i in self.variables:
                ret[i] = len(self.variables[i])
            return ret

    @property
    def is_hierarchical(self):
        return False


class CodeOutput(collections.abc.Mapping):

    __metaclass__ = ABCMeta

    def __init__(self):
        self.output_values = {}

    @abstractmethod
    def read(self):
        pass

    # True means that the run is just complete
    @property
    @abstractmethod
    def is_finished(self):
        pass

    @property
    def is_loaded(self):
        return not self.output_values == {}    

    def __contains__(self, x):
        return x in self.output_values

    def __getitem__(self, x):
        return self.output_values.__getitem__(x)
    
    def __iter__(self):
        return self.output_values.__iter__()

    def __len__(self):
        return self.output_values.__len__()




import time
from abc import ABCMeta, abstractmethod
from pychemia import pcm_log, HAS_PYMONGO

if HAS_PYMONGO:
    import bson


class Searcher:
    __metaclass__ = ABCMeta
    """
    General class for all optimization algorithms that uses fixed and blocked
    Generations
    """

    def __init__(self, population, generation_size=32, stabilization_limit=10,
                 target_value=None, searcher_id=None):
        self.population = population
        self.generation_size = generation_size
        self.stabilization_limit = stabilization_limit
        self.target_value = target_value
        self.current_generation = 0
        self.generation = {}
        self.sleep_time = 2
        self.delta_change = 0.2
        self.lineage = {}
        self.lineage_inv = {}

        self.old_actives = []
        self.old_nextgen = []
        self.pcdb = population.pcdb
        if searcher_id is None:
            self.searcher_id = self.population.tag
        else:
            self.searcher_id = searcher_id

    def recover(self, changedb=False):
        if self.pcdb is not None:
            data = self.pcdb.db.searcher_info.find_one({'_id': self.searcher_id})
            if data is not None:
                self.stabilization_limit = data['stabilization_limit']
                self.generation_size = data['generation_size']
                self.set_params(data['params'])
                self.generation = {}
                self.current_generation = 0

            for entry in self.get_all_generations():
                if self.searcher_id in entry:
                    self.generation[entry['_id']] = entry[self.searcher_id]
                    if max(self.generation[entry['_id']]) > self.current_generation:
                        self.current_generation = max(self.generation[entry['_id']])

            lineage_data = self.pcdb.db.lineage.find_one({'_id': self.searcher_id})
            if lineage_data is not None:
                lineage = lineage_data['lineage']
                lineage_inv = lineage_data['lineage_inv']

                changed = False
                for i in lineage.keys():
                    for j in lineage[i]:
                        if j == 'None':
                            lineage[i].remove(j)
                            changed = True
                if changed and changedb:
                    print('Lineage was changed (Some entry was null)')
                    self.pcdb.db.lineage.update_one({'_id': self.searcher_id}, {'$set': {'lineage': lineage}})

                sizes = [len(lineage[x]) for x in lineage]
                assert min(sizes) == max(sizes)

                for i in lineage.keys():
                    if HAS_PYMONGO:
                        self.lineage[i] = [bson.ObjectId(x) for x in lineage[i]]
                    else:
                        self.lineage[i] = [x for x in lineage[i]]

                for i in lineage_inv.keys():
                    if HAS_PYMONGO:
                        self.lineage_inv[bson.ObjectId(i)] = lineage_inv[i]
                    else:
                        self.lineage_inv[i] = lineage_inv[i]

                self.correct_extras(changedb)

    def correct_extras(self, changedb=False):
        for entry_id in self.get_generation():
            if entry_id not in self.lineage_inv:
                print('Disabling one entry not in lineage_inv', entry_id)
                self.population.disable(entry_id)
                print(self.generation.pop(entry_id))
                if changedb:
                    self.pcdb.db.generations.remove({'_id': entry_id})
            else:
                slot = self.lineage_inv[entry_id]
                if self.lineage[slot][-1] != entry_id:
                    print('Disabling one entry not in lineage[slot][-1]', entry_id)
                    self.population.disable(entry_id)
                    print(self.generation.pop(entry_id))
                    if changedb:
                        self.pcdb.db.generations.remove({'_id': entry_id})

        if self.current_generation > 0:
            for slot in range(self.generation_size):
                entry_id = self.lineage[str(slot)][-1]
                if entry_id not in self.population.actives:
                    print('Activating from lineage', entry_id)
                    self.population.enable(entry_id)
                if entry_id not in self.get_generation():
                    self.set_generation(entry_id, self.current_generation)
            actives = self.population.actives
            for entry_id in actives:
                if entry_id not in self.get_generation():
                    print('Disabling ', entry_id)
                    self.population.disable(entry_id)
            for entry_id in self.get_generation():
                if entry_id not in self.population.actives:
                    print('Enabling', entry_id)
                    self.population.enable(entry_id)
            candidates_per_generation = [len(self.get_generation(i)) for i in range(self.current_generation + 1)]
            pcm_log.info('Candidates per generation: %s' % candidates_per_generation)
            pcm_log.info('Current generation: %d Candidates: %d' % (self.current_generation,
                                                                    len(self.get_generation())))
            print('CANDIDATES:',candidates_per_generation)
            assert len(self.get_generation()) == self.generation_size
            assert min(candidates_per_generation) == max(candidates_per_generation)

    def __str__(self):
        ret = ' Searcher Name:       %s\n' % self.searcher_name
        ret += ' Generation size:     %d\n' % self.generation_size
        ret += ' Stabilization limit: %d\n' % self.stabilization_limit
        ret += ' Current Generation:  %d\n' % self.current_generation
        ret += ' Parameters: %s\n' % str(self.get_params())
        return ret

    def enforce_generation_size(self):

        while True:
            # Get a snapshot of the current status
            # Otherwise some inconsistencies could appear if some entries become evaluated during the execution
            # of this routine
            actives = self.population.actives
            evaluated = self.population.evaluated
            actives_evaluated = self.population.actives_evaluated
            actives_no_evaluated = self.population.actives_no_evaluated

            if len(actives) == 0:
                self.population.random_population(self.generation_size)
                break
            elif len(actives) == self.generation_size:
                pcm_log.debug('Equal')
                self.correct_extras(changedb=True)
                break
            elif len(actives) > self.generation_size:
                pcm_log.debug('More %d' % len(actives))
                self.recover(changedb=True)
                self.correct_extras(changedb=True)
            else:
                pcm_log.debug('Less %d' % len(actives))
                sizes = [len(self.lineage[x]) for x in self.lineage]
                assert min(sizes) == max(sizes)
                for i in self.lineage:
                    entry_id = self.lineage[i][-1]
                    print('Activating', i, entry_id)
                    self.population.enable(entry_id)
                assert len(self.population.actives) == self.generation_size
                assert len(self.get_generation()) == self.generation_size
                sizes = [len(self.get_generation(i)) for i in range(self.current_generation)]
                assert min(sizes) == max(sizes)

                # if len(evaluated) >= self.generation_size:
                #     pcm_log.debug('Enabling some entries starting from the best candidates')
                #     candidates = self.population.ids_sorted(evaluated)
                #     index = 0
                #     while len(actives) < self.generation_size and index < len(evaluated):
                #         if candidates[index] not in actives:
                #             pcm_log.debug('Not in actives %s' % str(candidates[index]))
                #             self.population.enable(candidates[index])
                #         else:
                #             pcm_log.debug('In actives %s' % str(candidates[index]))
                #         index += 1
                #         actives = self.population.actives
                # else:
                #     n = self.generation_size - len(actives)
                #     pcm_log.debug('Creating %d random structures' % n)
                #     self.population.random_population(n)

        for i in self.population.actives:
            self.set_generation(i, self.current_generation)

    def get_generation(self, generation_number=None):
        """
        Return all the elements tagged as belonging to a given generation

        :param generation_number: Number of generation to get
        :return: (list) Identifiers of all candidates for a given generation number
        :rtype: list
        """
        if generation_number is None:
            generation_number = self.current_generation
        from_gens = [x for x in self.generation if generation_number in self.generation[x]]
        ret = []
        for i in from_gens: 
            if i in self.lineage_inv:
                ret.append(i)
        return ret

    def print_status(self):
        pcm_log.info(' %s (tag: %s)' % (self.population.name, self.population.tag))
        pcm_log.info(' Current Generation             : %4d' % self.current_generation)
        pcm_log.info(' Population (evaluated/total)   : %4d /%4d' % (len(self.population.evaluated),
                                                                     len(self.population.members)))
        pcm_log.info(' Actives (evaluated/total)      : %4d /%4d' % (len(self.population.actives_evaluated),
                                                                     len(self.population.actives)))
        pcm_log.info(' Size of Generation (this/next) : %4d /%4d\n' % (len(self.get_generation()),
                                                                       len(self.get_generation(
                                                                         self.current_generation + 1))))

        if len(self.get_generation(self.current_generation + 1)) + len(self.population.actives) != self.generation_size:
            pcm_log.debug('Change in generations')
            for i in self.old_actives:
                if i not in self.population.actives:
                    pcm_log.debug('This active disappeared: %s' % str(i))
            for i in self.population.actives:
                if i not in self.old_actives:
                    pcm_log.debug('This active appeared: %s' % str(i))
            for i in self.old_nextgen:
                if i not in self.get_generation(self.current_generation + 1):
                    pcm_log.debug('This nextgen disappeared: %s' % str(i))
            for i in self.get_generation(self.current_generation + 1):
                if i not in self.old_nextgen:
                    pcm_log.debug('This nextgen appeared: %s' % str(i))
        self.old_actives = self.population.actives
        self.old_nextgen = self.get_generation(self.current_generation + 1)

    def advance(self, father, son, change):
        self.write_change(father, change)
        if father not in self.lineage_inv:
            print('%s not in current lineage' % father)
            print('Lineages %s' % self.lineage_inv.keys())
            raise ValueError('Father not in lineage')
        slot = self.lineage_inv[father]
        self.lineage[slot].append(None)
        self.lineage[slot][self.current_generation + 1] = son
        self.lineage_inv[son] = slot
        self.set_generation(son, self.current_generation + 1)

    def pass_to_new_generation(self, entry_id, reason=None):
        # pcm_log.debug('Moving to new generation : %s' % str(entry_id))
        change = {'change': 'promoted', 'reason': reason}
        self.advance(entry_id, entry_id, change)

    def replace_by_changed(self, entry_id_old, reason=None):
        entry_id_new = self.population.move_random(entry_id_old, factor=self.delta_change, in_place=False,
                                                   kind='change')
        change = {'change': 'modified', 'to': entry_id_new, 'reason': reason}
        pcm_log.debug('Modified  %s -> %s' % (str(entry_id_old), str(entry_id_new)))
        self.population.disable(entry_id_old)
        self.advance(entry_id_old, entry_id_new, change)

    def replace_by_other(self, entry_id_old, entry_id_new, reason=None):
        change = {'change': 'replace_by_other', 'to': entry_id_new, 'reason': reason}
        # pcm_log.debug('Changed  %s -> %s' % (str(entry_id_old), str(entry_id_new)))
        self.population.disable(entry_id_old)
        self.advance(entry_id_old, entry_id_new, change)

    def replace_by_random(self, entry_id_old, reason=None):
        self.population.disable(entry_id_old)
        entry_id_new, origin = self.population.add_random()
        pcm_log.debug('Replace by random  %s -> %s' % (str(entry_id_old), str(entry_id_new)))
        change = {'change': 'replace_by_random', 'to': entry_id_new, 'reason': reason, 'origin': origin}
        self.advance(entry_id_old, entry_id_new, change)

    def save_generations(self):
        if self.pcdb is not None:
            for entry_id in sorted(self.generation):
                info = self.generation[entry_id]
                if self.pcdb.db.generations.find_one({'_id': entry_id}) is None:
                    self.pcdb.db.generations.insert_one({'_id': entry_id, self.searcher_id: info})
                else:
                    self.pcdb.db.generations.update_one({'_id': entry_id}, {'$set': {self.searcher_id: info}})

            lineage = {}
            for i in self.lineage.keys():
                lineage[i] = [str(x) for x in self.lineage[i]]

            lineage_inv = {}
            for i in self.lineage_inv.keys():
                lineage_inv[str(i)] = self.lineage_inv[i]

            if self.pcdb.db.lineage.find_one({'_id': self.searcher_id}) is None:
                self.pcdb.db.lineage.insert_one({'_id': self.searcher_id,
                                             'lineage_inv': lineage_inv,
                                             'lineage': lineage})
            else:
                self.pcdb.db.lineage.update_one({'_id': self.searcher_id}, {'$set': {'lineage': lineage,
                                                                                 'lineage_inv': lineage_inv}})

    def set_generation(self, entry_id, value):
        if entry_id not in self.generation:
            self.generation[entry_id] = [value]
        else:
            if value not in self.generation[entry_id]:
                self.generation[entry_id].append(value)

    @abstractmethod
    def run_one(self):
        pass

    @abstractmethod
    def set_params(self, params):
        pass

    @abstractmethod
    def get_params(self):
        pass

    def update_lineages(self):

        assert len(self.population.actives) == self.generation_size

        # Complete lineage dictionary with generation_size entries
        # Each entry on the dictionary is a list of current_generation+1 values
        for i in range(self.generation_size):
            if str(i) not in self.lineage:
                self.lineage[str(i)] = [None]
            while len(self.lineage[str(i)]) < self.current_generation + 1:
                self.lineage[str(i)].append(None)

        # Complete lineage for all actives with known lineage_inv
        for entry_id in self.population.actives:
            if entry_id in self.lineage_inv:
                slot = self.lineage_inv[entry_id]
                self.lineage[slot][self.current_generation] = entry_id

        # Find empty slots for actives not in lineage_inv
        for entry_id in self.population.actives:
            if entry_id not in self.lineage_inv:
                for i in range(self.generation_size):
                    # Search for the first empty slot and associate it to the entry_id
                    if self.lineage[str(i)][self.current_generation] is None:
                        self.lineage[str(i)][self.current_generation] = entry_id
                        self.lineage_inv[entry_id] = str(i)
                        break

        # Check consistency between lineage and lineage_inv
        for i in range(self.generation_size):
            assert str(i) in self.lineage
            assert self.lineage[str(i)][-1] in self.lineage_inv
            assert self.lineage_inv[self.lineage[str(i)][-1]] == str(i)

        # Setting Generation list
        pcm_log.debug('Setting generation...')
        for entry_id in self.population.actives:
            self.set_generation(entry_id, self.current_generation)

    def run(self):
        """
        Execute the total number of cycles

        :return:
        """

        print("Population Information")
        print("======================")
        print(self.population)
        print("Searcher Information")
        print("====================")
        print(self)
        self.save_info()
        self.population.save_info()
        best_member = ''
        best_recorded = None
        survival_for_best = 0

        while True:
            print('\nGENERATION: %d' % self.current_generation)
            if len(self.population) > 0:
                self.print_status()

            pcm_log.info('[%s] Ensuring the number of active candidates is %d' % (self.searcher_name,
                                                                                  self.generation_size))
            self.enforce_generation_size()
            self.update_lineages()

            self.old_actives = self.population.actives
            self.old_nextgen = self.get_generation(self.current_generation + 1)

            self.print_status()

            number_evaluated = len(self.population.actives_evaluated)
            while self.population.fraction_evaluated < 1.0:
                if len(self.population.actives_evaluated) != number_evaluated:
                    msg = "Population '%s' still not evaluated. %4.0f %%"
                    pcm_log.debug(msg % (self.population.name, 100 * self.population.fraction_evaluated))
                    self.print_status()
                    number_evaluated = len(self.population.actives_evaluated)
                self.population.replace_failed()
                time.sleep(self.sleep_time)
            pcm_log.info("Population '%s' evaluated. %4.0f %%" % (self.population.name,
                                                                  100 * self.population.fraction_evaluated))

            best_member = self.population.best_candidate
            self.population.refine_progressive(best_member)

            pcm_log.info('Current best candidate: [%s]:\n%s' % (best_member, self.population.str_entry(best_member)))
            if best_member in self.get_generation():
                print('This candidate have survived for %d generations' % len(self.generation[best_member]))
                if len(self.generation[best_member]) >= self.stabilization_limit:
                    self.save_generations()
                    break
            else:
                pcm_log.debug('Best candidate %s is not in the current generation' % best_member)
                # pcm_log.debug('Slot: %s' % self.lineage_inv[best_member])
                # pcm_log.debug('Lineage: %s' % self.lineage[self.lineage_inv[best_member]])
                if best_member != best_recorded:
                    survival_for_best = 0
                    best_recorded = best_member
                else:
                    survival_for_best += 1

                if survival_for_best >= self.stabilization_limit:
                    print('This candidate have survived for %d generations' % survival_for_best)
                    self.save_generations()
                    break

            if self.target_value is not None:
                if self.population.value(best_member) <= self.target_value:
                    print('Target value achieved: target=%9.3f best=%9.3f' % (self.population.value(best_member),
                                                                              self.target_value))
                    self.save_generations()
                    break
                else:
                    print('Best value = %7.3f     target value = %7.3f' % (self.population.value(best_member),
                                                                           self.target_value))

            pcm_log.debug('[%s] Removing not evaluated: %d' %
                          (self.searcher_name, len(self.population.actives_no_evaluated)))
            for entry_id in self.population.actives_no_evaluated:
                self.replace_by_random(entry_id, reason='no evaluated')
            self.print_status()

            duplicates = self.population.get_duplicates(self.population.ids_sorted(self.population.actives_evaluated))
            print(duplicates)
            for entry_id in duplicates:
                change = {'change': 'duplicate', 'to': duplicates[entry_id], 'reason': None}
                self.write_change(entry_id, change)
                self.replace_by_random(entry_id, reason='duplicate')
            pcm_log.info(' Duplicates identified and disabled: %d' % len(duplicates))
            self.print_status()

            pcm_log.info(' Running one cycle for %s with %d candidates' % (self.searcher_name,
                                                                           len(self.actives_in_generation)))
            self.run_one()
            self.update_generation()

        print('Searcher ended after %d iterations' % self.current_generation)
        print('Best candidate: [%s]:\n%s' % (best_member, self.population.str_entry(best_member)))

    def write_change(self, entry_id, change):
        if self.pcdb is not None:
            change['method'] = self.searcher_name
            change['tag'] = self.searcher_name
            change['generation'] = self.current_generation
            change['from'] = entry_id
            self.pcdb.db.generation_changes.insert_one(change)

    def update_generation(self):
        # Increase the current generation number
        assert len(self.get_generation()) == self.generation_size
        self.current_generation += 1

        # Enable all entries in the new generation
        # Disable all entries not in new generation
        pcm_log.debug('[%s] Activating new generation, disabling others' % self.searcher_name)
        for entry_id in self.generation:
            if self.current_generation in self.generation[entry_id]:
                self.population.enable(entry_id)
            else:
                self.population.disable(entry_id)
        self.save_generations()

    @property
    def searcher_name(self):
        return self.__class__.__name__

    @property
    def to_dict(self):
        return {'name': self.searcher_name,
                'generation_size': self.generation_size,
                'stabilization_limit': self.stabilization_limit,
                'params': self.get_params()}

    def save_info(self):
        if self.pcdb is not None:
            data = self.pcdb.db.searcher_info.find_one({'_id': self.searcher_id})
            if data is None:
                data = self.to_dict
                data['_id'] = self.searcher_id
                self.pcdb.db.searcher_info.insert_one(data)
            else:
                self.pcdb.db.searcher_info.update_one({'_id': self.searcher_id}, self.to_dict)

    def clean(self):
        if self.pcdb is not None:
            self.pcdb.db.searcher_info.drop()
            self.pcdb.db.generations.drop()
            self.pcdb.db.generation_changes.drop()
            self.pcdb.db.lineage.drop()

    def replace_failed(self):
        self.pcdb.replace_failed()

    def get_all_generations(self, generation_number=None):
        if generation_number is None:
            return self.pcdb.db.generations.find()
        else:
            return self.pcdb.db.generations.find({self.population.tag: generation_number})

    @property
    def actives_in_generation(self):
        return [x for x in self.population.actives if self.population.is_evaluated(x) and x in self.get_generation()]



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import sys
import numpy as np
import pychemia
from pychemia.visual.searcher import *
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from pychemia.crystal import CrystalSymmetry
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection


def get_all_status(population):
    ret = []
    for entry in population.pcdb.entries.find({}, {'status': 1}):
        for x in entry['status']:
            if x not in ret:
                ret.append(x)
    return ret


def spacegroup2poly(spacegroup):
    """
    Return the size of polygon appropiated to represent a given
    crystal system

    :param spacegroup:
    :return:
    """
    if spacegroup <= 2:
        return 3
    elif spacegroup <= 15:
        return 5
    elif spacegroup <= 74:
        return 7
    elif spacegroup <= 142:
        return 9
    elif spacegroup <= 167:
        return 8
    elif spacegroup <= 194:
        return 6
    else:
        return 4


def label(ax, xy, spacegroup, energy):
    """
    Write the energy and spacegroup number on the center of each
    candidate polygon

    :param ax:
    :param xy:
    :param spacegroup:
    :param energy:
    :return:
    """
    y = xy[1] - 0.0  # shift y-value for label so that it's below the artist
    ax.text(xy[0], y, "%d" % spacegroup, ha="center", family='sans-serif', size=8)
    y = xy[1] - 0.2  # shift y-value for label so that it's below the artist
    ax.text(xy[0], y, "%.3f" % energy, ha="center", family='sans-serif', size=8)


def add_structure(ax, patches, position, spacegroup, energy):
    """
    Add one polygon to the patches list

    :param ax:
    :param patches:
    :param position:
    :param spacegroup:
    :param energy:
    :return:
    """
    polygon = mpatches.RegularPolygon(position, spacegroup2poly(spacegroup), 0.5, clip_on=True)
    patches.append(polygon)
    label(ax, position, spacegroup, energy)


def change_symbol(change):
    """
    Return the text that inform about the conditions that create that candidate

    :param change:
    :return:
    """

    if change == 'promoted':
        return 'P'
    elif change == 'modified':
        return 'M'
    elif change == 'replace_by_random':
        return 'RR'
    elif change == 'replace_by_other':
        return 'MV'
    elif change == 'duplicate':
        return '___'
    else:
        print(change)


def get_generation(searcher, tag):
    """
    Return the identifiers of all members of a population associated with
    a given searcher tag

    :param searcher:
    :param tag:
    :return:
    """
    lista = [x['_id'] for x in searcher.get_all_generations(tag) if searcher.population.is_evaluated(x['_id'])]
    return searcher.population.ids_sorted(lista)


def get_generation_limits(searcher, gen_size):
    inigen = 0
    fingen = 0
    i = 0
    while True:
        n = len(get_generation(searcher, i))
        print(' [Generation %d: Number candidates = %d]' % (i, n))
        if n < gen_size:
            fingen = i
            break
        i += 1
    return inigen, fingen


def plot_generation_chart(searcher, gen_size):
    colors = []
    patches = []
    population = searcher.population

    inigen, fingen = get_generation_limits(searcher, gen_size)

    if inigen == fingen:
        return

    fig, ax = plt.subplots(figsize=(gen_size, 2 * (fingen - inigen)))

    avg_energies = np.zeros(fingen)

    # Structures
    sys.stdout.write("Plotting Structures for generations:")
    for j in range(inigen, fingen):
        sys.stdout.write(' %d' % j)
        thegen = get_generation(population, j)[:gen_size]
        plt.text(-0.9, -2 * j, "Gen %d" % j, ha="center", family='sans-serif', size=12)
        energies = []
        for i in range(gen_size):
            struct, properties, status = population.pcdb.get_dicts(thegen[i])
            # print properties['spacegroup'], properties['energy_pa']
            spacegroup = properties['spacegroup']
            energy = properties['energy_pa']
            energies.append(energy)
            add_structure(ax, patches, [i, -2 * j], spacegroup, energy)
            colors.append(energy)
        avg_energies[j] = np.mean(energies[:navg_energy])

    collection = PatchCollection(patches, cmap=matplotlib.cm.jet, alpha=0.3)
    collection.set_array(np.array(colors))
    ax.add_collection(collection)
    sys.stdout.write('\n')

    # Duplicates
    sys.stdout.write("Marking duplicates for generations:")
    for j in range(inigen, fingen):
        sys.stdout.write(' %d' % j)
        thegen = get_generation(population, j)[:gen_size]
        dups = population.get_duplicates(thegen)
        for i in dups:
            x = [list(thegen).index(i), list(thegen).index(dups[i])]
            y = [-2 * j + 0.5, -2 * j + 0.5]
            plt.plot(x, y, 'k-')
    sys.stdout.write('\n')

    # Changes
    sys.stdout.write("Plotting changes for generations:")
    for j in range(inigen, fingen - 1):
        sys.stdout.write(' %d' % j)
        sys.stdout.flush()
        next_generation = get_generation(population, j + 1)
        thegen = get_generation(population, j)[:gen_size]
        for i in range(gen_size):
            # add a line
            for entry in population.pcdb.db.generation_changes.find({'from': thegen[i]}):
                csymbol = change_symbol(entry['change'])
                if csymbol is None:
                    print(entry)
                plt.text(i, -2 * j - 0.6, csymbol, ha="center", family='sans-serif', size=6)
                if 'to' in entry and entry['to'] in next_generation[:gen_size]:
                    next_id = list(next_generation).index(entry['to'])
                    x, y = np.array([[i, next_id], [-2 * j - 0.7, -2 * j - 1.3]])
                    line = mlines.Line2D(x, y, lw=3., alpha=0.3)
                    ax.add_line(line)
                elif entry['change'] == 'promoted' and entry['from'] in next_generation[:gen_size]:
                    next_id = list(next_generation[:gen_size]).index(entry['from'])
                    x, y = np.array([[i, next_id], [-2 * j - 0.7, -2 * j - 1.3]])
                    line = mlines.Line2D(x, y, lw=3., alpha=0.3)
                    ax.add_line(line)
    sys.stdout.write('\n')

    plt.title('%s (tag=%s)' % (population.name, population.tag))
    plt.text(float(gen_size) / 2, 1, '%s (tag=%s)' % (population.name, population.tag),
             ha="center",
             family='sans-serif',
             size=16)
    plt.subplots_adjust(left=0, right=1, bottom=0, top=1)
    plt.axis('equal')
    plt.axis('off')
    pdf.savefig()
    plt.clf()
    plt.close()

    # Average energy of best structures
    print('Plot: Average energy of best structures')
    plt.figure(figsize=letterpage)
    plt.plot(range(fingen), avg_energies, 'ro')
    plt.plot(range(fingen), avg_energies, 'b--')
    plt.xlabel('Generation')
    plt.ylabel('Average energy of %d best structures' % navg_energy)
    plt.title('%s (tag=%s)' % (population.name, population.tag))
    plt.xticks(range(fingen))
    pdf.savefig()
    plt.close()


def plot_evolution_circular(searcher, target_function='energy_pa', tag='spacegroup'):
    fig, ax = plt.subplots(figsize=(10, 10))
    ax.cla()
    plt.subplots_adjust(left=0.0, bottom=0.0, right=1.0, top=1.0, wspace=None, hspace=None)
    ax.set_xlim((-1.1, 1.1))
    ax.set_ylim((-1.1, 1.1))
    ax.set_axis_off()
    ax.set_aspect(1.0)

    for gen in range(searcher.current_generation):
        radius = 1.0 / (1.0 + np.exp(-6 * float(gen + 1) / searcher.current_generation))
        radius = (float(gen + 1) / searcher.current_generation) ** 2
        entries = {}
        counter = 0
        nele = len(searcher.lineage)
        for lin in range(nele):
            entry = searcher.population.get_entry(searcher.lineage[str(lin)][gen])
            _id = entry['_id']
            entries[_id] = {}
            entries[_id][target_function] = entry['properties'][target_function]
            if tag is not None:
                if tag in entry['properties']:
                    entries[_id][tag] = entry['properties'][tag]
                elif tag in entry['structure']:
                    entries[_id][tag] = entry['structure'][tag]
                elif tag == 'spacegroup':
                    st = searcher.population.get_structure(_id)
                    ss = CrystalSymmetry(st)
                    entries[_id][tag] = ss.number(1E-3)
                else:
                    raise ValueError('Tag not found')
            entries[_id]['x'] = np.cos(counter * 2 * np.pi / nele)
            entries[_id]['y'] = np.sin(counter * 2 * np.pi / nele)
            counter += 1

        ids = [i for i in entries]
        xx = np.array([entries[i]['x'] for i in ids])
        yy = np.array([entries[i]['y'] for i in ids])
        aa = np.array([4E3 for i in ids])
        cc = np.array([entries[i][target_function] for i in ids])
        best_candidate = entries[ids[cc.argsort()[0]]]
        plt.scatter(radius * xx, radius * yy, s=radius ** 2 * aa, c=cc, alpha=0.5 * radius, edgecolors='none')
        if tag is not None:
            for i in entries:
                if i == best_candidate:
                    ax.text(radius * entries[i]['x'], radius * entries[i]['y'], str(entries[i][tag]),
                            va='center',
                            ha='center', size=int(radius * 24), color='red')
                else:
                    ax.text(radius * entries[i]['x'], radius * entries[i]['y'], str(entries[i][tag]),
                            va='center',
                            ha='center', size=int(radius * 14))
    plt.savefig(searcher.population.name + '_circular.png')


gen_size = 80
letterpage = (11, 8.5)

dbsettings = {'host': 'mongo01.systems.wvu.edu', 'name': 'OrbitalGAF', 'user': 'gufranco', 'passwd': 'zxcvbnm'}
pcdb = pychemia.db.get_database(dbsettings)
popu = pychemia.population.orbitaldftu.OrbitalDFTU(pcdb)
searcher = pychemia.searcher.FireFly(popu, generation_size=80)
searcher.recover()

if __name__ == '__main__':
    pdf = PdfPages(name + 'multipage.pdf')

    print(popu)

    colors = ['r', 'g', 'm', 'c', 'y', 'k']
    color_tags = {}
    for i in tags:
        if len(colors) == 0:
            colors = ['r', 'g', 'm', 'c', 'y', 'k']
        color_tags[i] = colors.pop(0)

    plot_generation_chart(searcher, gen_size)

    sys.exit(1)

    # Change of energy vs distance
    print('Plot: Change of energy vs distance')
    ret = {}
    color = {'replace_by_other': 'r', 'replace_by_random': 'b'}
    plt.figure(figsize=letterpage)
    for change in ['replace_by_other', 'replace_by_random']:
        print('Searching for: ', change)
        ret[change] = []
        for entry in popu.pcdb.db.generation_changes.find({'change': change}):
            id_from = entry['from']
            id_to = entry['to']
            # print "%s --> %s" % (id_from, id_to)
            x = popu.distance(id_from, id_to)
            if popu.is_evaluated(id_from) and popu.is_evaluated(id_to):
                values = popu.get_values([id_from, id_to])
                delta_energy = values[id_to] - values[id_from]
                y = delta_energy
                ret[change].append([x, y])
        ret[change] = np.array(ret[change])
        if len(ret[change]) > 0:
            plt.plot(ret[change][:, 0], ret[change][:, 1], color[change] + 'o', label=change)

    plt.xlim(min(ret[change][:, 0]) - 0.01, 0.01 + max(ret[change][:, 0]))
    plt.xlabel('Distance between structures (after relaxation)')
    plt.ylabel('Energy change')
    plt.title('Change of energy due to movement of structures')
    plt.legend()
    pdf.savefig()
    plt.close()

    # Best energies vs Generations
    miny = min(known_structures['energy_pa'])
    maxy = max(known_structures['energy_pa'])
    rangey = (miny - 0.1 * (maxy - miny), maxy + 0.1 * (maxy - miny))

    plt.figure(figsize=letterpage)

    maxx = 0
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)

        inigen = fingen = 0
        inigen, fingen = get_generation_limits(popu, gen_size)

        if fingen > maxx:
            maxx = fingen

        # All energies vs Generations
        sys.stdout.write("Plot: All energies vs Generations:")
        for j in range(inigen, fingen):
            sys.stdout.write(' %d' % j)
            thegen = get_generation(popu, j)
            energies = []
            for i in range(len(thegen)):
                struct, properties, status = popu.pcdb.get_dicts(thegen[i])
                energy = properties['energy_pa']
                energies.append(energy)
            if len(energies) > 0:
                for i in range(len(energies)):
                    plt.plot([j, j + 1 - 0.1], [energies[i], energies[i]], color_tags[itag] + '-')
                    plt.text(j + 0.9 * float(i) / len(energies), energies[i], str(i),
                             ha="center", family='sans-serif', size=6)
        sys.stdout.write('\n')

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([inigen, maxx], energy_pa * np.ones(2), 'b--')
        plt.text(fingen, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    plt.xticks(range(maxx))
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Generations')
    plt.title('Energies')
    pdf.savefig()
    plt.close()

    plt.figure(figsize=letterpage)
    maxx = 0
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)

        inigen, fingen = get_generation_limits(popu, gen_size)
        if fingen > maxx:
            maxx = fingen

        sys.stdout.write("Plot: Best energies vs Generations:")
        best = 16
        for j in range(inigen, fingen):
            sys.stdout.write(' %d' % j)
            thegen = get_generation(popu, j)[:best]
            energies = []
            spacegroups = []
            for i in range(len(thegen)):
                struct, properties, status = popu.pcdb.get_dicts(thegen[i])
                energies.append(properties['energy_pa'])
                spacegroups.append(properties['spacegroup'])
            if len(energies) > 0:
                for i in range(len(energies)):
                    if rangey[0] < energies[i] < rangey[1]:
                        plt.plot([j, j + 1 - 0.1], [energies[i], energies[i]], color_tags[itag] + '-')
                        plt.text(j + 0.9 * float(i) / len(energies), energies[i], str(i),
                                 ha="center", family='sans-serif', size=6, color='0.9')
                        plt.text(j + 0.9, energies[i], str(spacegroups[i]),
                                 ha="center", family='sans-serif', size=6, color=color_tags[itag])
        sys.stdout.write('\n')

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([inigen, maxx], energy_pa * np.ones(2), 'b--')
        plt.text(maxx, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    plt.ylim(*rangey)
    plt.xticks(range(maxx))
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Generations')
    plt.title('Energies')
    pdf.savefig()

    plt.figure(figsize=letterpage)

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([0, 1], energy_pa * np.ones(2), 'b--')
        plt.text(0.5, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    index = 1
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)
        pre_energy = 1E10

        for entry_id in popu.ids_sorted(popu.evaluated):
            entry = popu.get_entry(entry_id)

            if itag in entry['status']:
                energy_pa = entry['properties']['energy_pa']
                spacegroup = entry['properties']['spacegroup']
                if np.abs(pre_energy - energy_pa) > 1E-3 and rangey[0] < energy_pa < rangey[1]:
                    plt.plot([index, index + 1], energy_pa * np.ones(2), color_tags[itag] + '--')
                    plt.text(index + 0.5, energy_pa, str(spacegroup),
                             ha="left",
                             family='sans-serif',
                             size=6,
                             color=color_tags[itag])
                    pre_energy = energy_pa

        index += 1

    plt.ylim(*rangey)
    plt.xticks(0.5 + np.arange(len(tags) + 1), ['CIFs'] + tags)
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Structures found')
    plt.title('Energies')
    pdf.savefig()

    plt.figure(figsize=letterpage)

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([0, 1], energy_pa * np.ones(2), 'b-')
        # plt.text(0.5, energy_pa,str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
    popu = pychemia.population.RelaxStructures(pcdb, distance_tol=0.4, value_tol=0.1)
    pre_energy = 1E10

    selection = popu.evaluated
    n = len(selection)
    while True:
        selection = popu.cleaned_from_duplicates(selection)
        if len(selection) == n:
            break
        n = len(selection)
        print('Selection', n)

    for entry_id in popu.ids_sorted(selection):
        entry = popu.get_entry(entry_id)
        energy_pa = entry['properties']['energy_pa']
        spacegroup = entry['properties']['spacegroup']
        if np.abs(pre_energy - energy_pa) > 1E-3 and rangey[0] < energy_pa < rangey[1]:
            plt.plot([1, 2], energy_pa * np.ones(2), 'r-')
            # plt.text(1.5, energy_pa,str(spacegroup), ha="left", family='sans-serif', size=6, color='r')
            pre_energy = energy_pa

    plt.ylim(*rangey)
    plt.xticks(0.5 + np.arange(2), ['Known Structures', 'New predicted'])
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Structures found')
    plt.title('Energies')
    pdf.savefig()

    pdf.close()



"""
Modules to manipulate input and output for several atomistic simulation codes.
Currently there are implementations for *ABINIT*, *DFTB+*, *Fireball*, an internal
LennardJones 'calculator', *Octopus* and *VASP*.

"""
from .relaxator import Relaxator
from . import vasp
from . import dftb
from . import lennardjones
from . import fireball
from . import sprkkr
from . import phonopy
from . import new_fireball
from . import siesta
from . import elk
from . import abinit
from . import octopus

# __all__ = filter(lambda s: not s.startswith('_'), dir())



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import sys
import logging
import pychemia
import subprocess
import xml.etree.ElementTree as ElementTree

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('pychemia')
logger.addHandler(logging.NullHandler())
logger.setLevel(logging.DEBUG)

version = 0.1


def help_info():
    print(""" pychemia_launcher.py
Search for suitable directories for relaxation

   Use:

       pychemia_launcher.py --user user_name
                            --queue queue_name
                            --mail mail_job_info
                            --path where_to_search_for_structures
                            [--binary name_of_executable]
   """)


def find_structures(basedir, filename='to_relax'):
    ret = []
    if filename in os.listdir(basedir):
        ret.append(basedir)
    for i in [x for x in os.listdir(basedir) if os.path.isdir(basedir + os.sep + x)]:
        ret += find_structures(basedir + os.sep + i)
    return ret


def get_structure_file(basedir):
    files = os.listdir(basedir)
    for ifile in ['structure_phase2.json', 'structure_phase1.json', 'structure2.json',
                  'structure1.json', 'structure.json', 'POSCAR']:
        if ifile in files:
            return ifile
    return None


if __name__ == '__main__':

    # Script starts from here
    if len(sys.argv) < 2:
        help_info()
        sys.exit(1)

    user = None
    mail = None
    queue = None
    path = None
    binary = 'vasp'
    nparal = 4
    nhours = 24

    for istruct in range(1, len(sys.argv)):
        if sys.argv[istruct].startswith('--'):
            option = sys.argv[istruct][2:]
            # fetch sys.argv[1] but without the first two characters
            if option == 'version':
                print(version)
                sys.exit()
            elif option == 'help':
                help_info()
                sys.exit()
            elif option == 'user':
                user = sys.argv[istruct + 1]
            elif option == 'mail':
                mail = sys.argv[istruct + 1]
            elif option == 'queue':
                queue = sys.argv[istruct + 1]
            elif option == 'path':
                path = sys.argv[istruct + 1]
            elif option == 'binary':
                binary = sys.argv[istruct + 1]
            elif option == 'nparal':
                nparal = int(sys.argv[istruct + 1])
            elif option == 'nhours':
                nhours = int(sys.argv[istruct + 1])
            else:
                print('Unknown option. --' + option)

    if path is None or user is None or queue is None:
        help_info()

    data = subprocess.check_output(['qstat', '-x', '-f', '-u', user])
    xmldata = ElementTree.fromstring(data)
    jobs = [i.find('Job_Name').text for i in xmldata.findall('Job')]

    found_structures = find_structures(path)

    for istruct in found_structures:
        if os.path.isfile(istruct + os.sep + 'lock'):
            print("Locked:    %s" % istruct)
        elif os.path.basename(istruct) in jobs:
            print('Submitted: %s' % istruct)
        else:
            print('To submit: %s' % istruct)
            rr = pychemia.runner.PBSRunner(istruct)
            rr.initialize(nodes=1, ppn=nparal, mail=mail, queue=queue, walltime=[0, nhours, 0, 0])
            structure_file = get_structure_file(istruct)
            if structure_file is None:
                print('No suitable structure was found on: ', istruct)
            rr.set_template('pcm_vasp_relaxator.py --binary %s --nparal %d --structure_file %s'
                            % (binary, nparal, structure_file))
            rr.write_pbs()
            rr.submit()



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import sys
import numpy as np
from pychemia.utils.constants import bohr_angstrom
import pychemia.code.abinit


def input2xyz(abivar, basename, datasets):
    """
    Take an input object 'av' and write the
    corresponding XYZ file in the given
    'basename'
    """

    for idts in datasets:
        struct = abivar.get_structure(idts)
        natom = struct.natom
        if idts == '':
            filep = basename + '.xyz'
        else:
            filep = basename + '_DS' + str(idts) + '.xyz'
        wf = open(filep, 'w')
        wf.write(str(natom) + '\n\n')
        for iatom in range(natom):
            wf.write('%s %16.8f %16.8f %16.8f\n' % (struct.symbols[iatom],
                                                    struct.positions[iatom, 0],
                                                    struct.positions[iatom, 1],
                                                    struct.positions[iatom, 2]))
        wf.close()


def write_one(time, nchist, wf, natom, struct):
    xcart = np.array(nchist['xcart'][time])
    atom_pos = bohr_angstrom * xcart
    wf.write(str(natom) + '\n\n')
    for iatom in range(natom):
        wf.write('%s %16.8f %16.8f %16.8f\n' % (struct['atom_names'][iatom], atom_pos[iatom, 0], atom_pos[iatom, 1],
                                                atom_pos[iatom, 2]))


def abihist2xyz(abivar, basename, datasets, time='all'):
    """
    Write XYZ files for history
    """
    for idts in datasets:

        # Reading the history
        if idts == '':
            history = abifile.files['tmpout'] + "_HIST"
        else:
            history = abifile.files['tmpout'] + "_DS" + str(idts) + "_HIST"
        if os.path.isfile(history):
            print('Reading ', history)
            nchist = pychemia.code.abinit.netcdf2dict(history)

            # Setting the output file
            if time == 'all':
                if idts == '':
                    filep = basename + '_HIST.xyz'
                else:
                    filep = basename + '_DS' + str(idts) + '_HIST.xyz'
            else:
                if idts == '':
                    filep = basename + '_HIST_itime' + str(time) + '.xyz'
                else:
                    filep = basename + '_DS' + str(idts) + '_HIST_itime' + str(time) + '.xyz'

            print('Writing ', filep)
            wf = open(filep, 'w')

            # Getting the atomic structure
            struct = abivar.get_struct(idts)
            ntime = len(nchist['mdtime'])
            natom = struct['natom']

            # Write the xyz section
            if time == 'all':
                sys.stdout.write('[')
                for it in range(ntime):
                    sys.stdout.write('*')
                    write_one(it, nchist, wf, natom, struct)
                sys.stdout.write(']\n')
            else:
                write_one(time, nchist, wf, natom, struct)

            # Close the file
            wf.close()


def helper():
    print('''\
   Extract atomic coordinates from abinit files
   This program can deal with the following
   files:
           abinit.files ('.files': abinit format)
           abinit.in    ('.in': input for abinit)
           *_OUT.nc     ('_OUT.nc': output in NetCDF file format)

   Options include:
     --version : Prints the version number
     --help    : Display this help
     --dtset # : Extract from a given dataset
     --itime # : Extract only a given iteration (--hist mandatory)
     --hist    : Extract from HIST file
     --input   : Extract from input file (default)
     --output  : Extract from output file
     ''')


def get_input_dts(filep, dataset):
    abivar = pychemia.code.abinit.AbinitInput(filep)
    datasets = None
    if dataset == 0:
        datasets = abivar.get_dtsets_keys()
    elif not str(dataset) in abivar.get_dtsets_keys():
        print('ERROR No such dataset index:', dataset)
        print('Valid values are:', abivar.get_dtsets_keys())
        exit(1)
    else:
        datasets = [dataset]
    return abivar, datasets


if __name__ == '__main__':

    # Script starts from here
    if len(sys.argv) < 2:
        helper()
        sys.exit(1)

    filename = ''
    dtset = 0
    itime = 'all'
    has_input = True
    has_output = False
    hist = False
    for i in range(1, len(sys.argv)):
        if sys.argv[i].startswith('--'):
            option = sys.argv[i][2:]
            # fetch sys.argv[1] but without the first two characters
            if option == 'version':
                print('Version 1.0')
                sys.exit()
            elif option == 'help':
                helper()
                sys.exit()
            elif option == 'dtset':
                dtset = int(sys.argv[i + 1])
            elif option == 'itime':
                itime = int(sys.argv[i + 1])
            elif option == 'input':
                has_input = True
                has_output = False
            elif option == 'output':
                has_output = True
                has_input = False
            elif option == 'hist':
                has_output = False
                has_input = False
                hist = True
            else:
                print('Unknown option. --' + option)

    filename = sys.argv[-1]
    if not os.path.isfile(filename):
        print('No valid filename')
        helper()
        sys.exit(1)

    av = None
    dts = None
    if filename != '':
        if filename[-6:] == '.files':
            abifile = pychemia.code.abinit.AbiFiles(filename)
            if has_input or hist:
                (av, dts) = get_input_dts(abifile.get_input_filename(), dtset)
            elif has_output:
                (av, dts) = get_input_dts(abifile.get_out_filename(), dtset)
            xyz = filename[:-6]
            if not hist:
                input2xyz(av, xyz, dts)
            else:
                abihist2xyz(av, xyz, dts, itime)
        elif filename[-3:] == '.in':
            (av, dts) = get_input_dts(filename, dtset)
            xyz = filename[:-3]
            input2xyz(av, xyz, dts)
        elif filename[-7:] == '_OUT.nc':
            (av, dts) = get_input_dts(filename, dtset)
            xyz = filename[:-7]
            input2xyz(av, xyz, dts)
        else:
            print('Filename with unknown extension')
    else:
        print('No file specified.')
        helper()
        sys.exit(1)
